# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import absolute_import, division, print_function

import base64
import binascii
import os
import struct
import time

import six

from cryptography import utils
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.hmac import HMAC


class InvalidToken(Exception):
    pass


_MAX_CLOCK_SKEW = 60


class Fernet(object):
    def __init__(self, key, backend=None):
        if backend is None:
            backend = default_backend()

        key = base64.urlsafe_b64decode(key)
        if len(key) != 32:
            raise ValueError(
                "Fernet key must be 32 url-safe base64-encoded bytes."
            )

        self._signing_key = key[:16]
        self._encryption_key = key[16:]
        self._backend = backend

    @classmethod
    def generate_key(cls):
        return base64.urlsafe_b64encode(os.urandom(32))

    def encrypt(self, data):
        current_time = int(time.time())
        iv = os.urandom(16)
        return self._encrypt_from_parts(data, current_time, iv)

    def _encrypt_from_parts(self, data, current_time, iv):
        utils._check_bytes("data", data)

        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        padded_data = padder.update(data) + padder.finalize()
        encryptor = Cipher(
            algorithms.AES(self._encryption_key), modes.CBC(iv), self._backend
        ).encryptor()
        ciphertext = encryptor.update(padded_data) + encryptor.finalize()

        basic_parts = (
            b"\x80" + struct.pack(">Q", current_time) + iv + ciphertext
        )

        h = HMAC(self._signing_key, hashes.SHA256(), backend=self._backend)
        h.update(basic_parts)
        hmac = h.finalize()
        return base64.urlsafe_b64encode(basic_parts + hmac)

    def decrypt(self, token, ttl=None):
        timestamp, data = Fernet._get_unverified_token_data(token)
        return self._decrypt_data(data, timestamp, ttl)

    def extract_timestamp(self, token):
        timestamp, data = Fernet._get_unverified_token_data(token)
        # Verify the token was not tampered with.
        self._verify_signature(data)
        return timestamp

    @staticmethod
    def _get_unverified_token_data(token):
        utils._check_bytes("token", token)
        try:
            data = base64.urlsafe_b64decode(token)
        except (TypeError, binascii.Error):
            raise InvalidToken

        if not data or six.indexbytes(data, 0) != 0x80:
            raise InvalidToken

        try:
            timestamp, = struct.unpack(">Q", data[1:9])
        except struct.error:
            raise InvalidToken
        return timestamp, data

    def _verify_signature(self, data):
        h = HMAC(self._signing_key, hashes.SHA256(), backend=self._backend)
        h.update(data[:-32])
        try:
            h.verify(data[-32:])
        except InvalidSignature:
            raise InvalidToken

    def _decrypt_data(self, data, timestamp, ttl):
        current_time = int(time.time())
        if ttl is not None:
            if timestamp + ttl < current_time:
                raise InvalidToken

            if current_time + _MAX_CLOCK_SKEW < timestamp:
                raise InvalidToken

        self._verify_signature(data)

        iv = data[9:25]
        ciphertext = data[25:-32]
        decryptor = Cipher(
            algorithms.AES(self._encryption_key), modes.CBC(iv), self._backend
        ).decryptor()
        plaintext_padded = decryptor.update(ciphertext)
        try:
            plaintext_padded += decryptor.finalize()
        except ValueError:
            raise InvalidToken
        unpadder = padding.PKCS7(algorithms.AES.block_size).unpadder()

        unpadded = unpadder.update(plaintext_padded)
        try:
            unpadded += unpadder.finalize()
        except ValueError:
            raise InvalidToken
        return unpadded


class MultiFernet(object):
    def __init__(self, fernets):
        fernets = list(fernets)
        if not fernets:
            raise ValueError(
                "MultiFernet requires at least one Fernet instance"
            )
        self._fernets = fernets

    def encrypt(self, msg):
        return self._fernets[0].encrypt(msg)

    def rotate(self, msg):
        timestamp, data = Fernet._get_unverified_token_data(msg)
        for f in self._fernets:
            try:
                p = f._decrypt_data(data, timestamp, None)
                break
            except InvalidToken:
                pass
        else:
            raise InvalidToken

        iv = os.urandom(16)
        return self._fernets[0]._encrypt_from_parts(p, timestamp, iv)

    def decrypt(self, msg, ttl=None):
        for f in self._fernets:
            try:
                return f.decrypt(msg, ttl)
            except InvalidToken:
                pass
        raise InvalidToken



import json

from application.interactor.unlock_voucher import UnlockVoucherInteractor
from application.interactor.unlock_voucher.dto import UnlockVoucherRequest
from application.model import AsyncUuidMessageProcessor
from messagebus import (
    DefaultToken,
    Message,
    InternalErrorMessage,
)


class UnlockVoucherProcessor(AsyncUuidMessageProcessor):
    def __init__(
        self,           # type: UnlockVoucherProcessor
        interactor,     # type: UnlockVoucherInteractor
    ):
        # type: (...) -> None

        super(UnlockVoucherProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(
        self,           # type: UnlockVoucherProcessor
        data,           # type: str
        event_type,     # type: str
    ):
        # type: (str, str) -> UnlockVoucherRequest

        request_json = json.loads(data.decode("utf-8"))
        return UnlockVoucherRequest(
            benefit_id=request_json["benefitId"],
            order_picture=request_json["orderPicture"],
            table_picture=request_json.get("tablePicture")
        )

    def call_business(
        self,           # type: UnlockVoucherProcessor
        request,        # type: UnlockVoucherRequest
    ):
        # type: (...) -> None

        self.interactor.unlock_benefit(
            request=request,
        )

    def format_response(self, message_bus, message, event, request, response):
        message_bus.reply_message(
            message=message,
            reply_message=Message(token=DefaultToken.TK_SYS_ACK)
        )

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(
            message=message,
            reply_message=InternalErrorMessage(ex=exception)
        )

    def format_parse_exception(self, message_bus, message, event, data, exception):
        self.format_exception(
            message_bus=message_bus,
            message=message,
            event=event,
            input_obj=data,
            exception=exception
        )



from typing import (
    List,
    Optional,
)


class TableReportSaleLineDto(object):

    def __init__(self):
        self.product_name = ""                  # type: str
        self.price = 0.0                        # type: float
        self.qty = 0                            # type: float
        self.default_qty = 0                    # type: Optional[int]
        self.item_type = ""                     # type: str
        self.seat = None                        # type: Optional[int]
        self.level = 0                          # type: int
        self.comment = ""                       # type: Optional[str]
        self.line_number = 0                    # type: int
        self.service_period = None              # type: Optional[str]
        self.order_id = None                    # type: Optional[int]
        self.item_id = None                     # type: Optional[str]
        self.part_code = None                   # type: Optional[int]
        self.overwritten_unit_price = None      # type: Optional[float]
        self.fraction_parent = None             # type: Optional[int]
        self.original_related_line_ids = None   # type: Optional[List[str]]



import json

from application.interactor import LockBenefitInteractor
from application.model import AsyncUuidMessageProcessor
from application.model.customexception import ClientCustomError, NotFoundError, InUseError, GoneError
from application.processor.get_and_lock_coupon.dto import GetAndLockCouponErrorCode
from messagebus import (
    DefaultToken,
    DataType,
    Message,
    MessageBus,
    AckMessage,
    Event,
    InternalErrorMessage,
)
from typing import (
    Any,
    Optional,
)
from ..tokens import ResponseToken


class LockBenefitProcessor(AsyncUuidMessageProcessor):
    def __init__(self, interactor):
        # type: (LockBenefitInteractor) -> None

        super(LockBenefitProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(self, data, event_type):
        # type: (str, str) -> Any

        parsed_data = json.loads(data)
        benefit_id = str(parsed_data["benefitId"])
        pos_id = int(parsed_data["posId"])
        return benefit_id, pos_id

    def call_business(self, params):
        # type: ((int, str)) -> None

        benefit_id, pos_id = params

        self.interactor.lock_benefit(
            benefit_id=benefit_id,
            pos_id=pos_id,
        )

    def format_response(
        self,
        message_bus,  # type: MessageBus
        message,      # type: Message
        event,        # type: Optional[Event]
        request,      # type: Optional[Any]
        response,     # type: str
    ):
        message_bus.reply_message(
            message=message,
            reply_message=AckMessage(response)
        )

    def format_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Event]
        input_obj,      # type: Optional[Any]
        exception,      # type: Exception
    ):
        # type: (...) -> None

        data = ""
        if isinstance(exception, ClientCustomError):
            data = json.dumps({
                "message": {
                    "pt": exception.message.pt,
                    "en": exception.message.en,
                    "es": exception.message.es,
                }
            }).encode("utf-8")
            error_code = GetAndLockCouponErrorCode.PARTNER_CUSTOM_ERROR
        elif isinstance(exception, NotFoundError):
            error_code = GetAndLockCouponErrorCode.COUPON_NOT_FOUND
        elif isinstance(exception, InUseError):
            error_code = GetAndLockCouponErrorCode.COUPON_LOCKED
        elif isinstance(exception, GoneError):
            error_code = GetAndLockCouponErrorCode.COUPON_ALREADY_USED
        else:
            message_bus.reply_message(
                message=message,
                reply_message=InternalErrorMessage(ex=exception)
            )
            return

        message_bus.reply_message(
            message=message,
            reply_message=Message(
                token=ResponseToken.TK_LOYALTY_ERROR.value,
                data_type=DataType.param,
                data=error_code.name + "\0" + data,
            ),
        )

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



from typing import Dict

from justificationdiscountreport.repository import JustificationDiscountRepository


class JustificationDiscountReport(object):
    def __init__(
        self,
        justification_discount_repository,   # type: JustificationDiscountRepository
    ):
        # type: (...) -> None

        self.justification_discount_repository = justification_discount_repository

    def get_justifications(
        self,
        initial_date,       # type: str
        end_date,           # type: str
        operator_id,        # type: str
    ):
        # type: (...) -> Dict

        justifications = self.justification_discount_repository.get_justifications(
            initial_date=initial_date,
            end_date=end_date,
            operator_id=operator_id,
        )

        result = {}
        list_order_date = list(dict.fromkeys([item.order_date for item in justifications]))
        for order_date in list_order_date:
            filtered_justifications = filter(
                (lambda items: items.order_date == order_date),
                justifications,
            )

            if len(filtered_justifications) > 0:
                result[str(order_date)] = filtered_justifications

        return result



from application.model import Configurations
from application.service import BenefitService
from application.service import OrderService


class OnOrderTotalInteractor(object):
    def __init__(self, configs, benefit_service, order_service):
        # type: (Configurations, BenefitService, OrderService) -> None

        self.configs = configs
        self.logger = configs.logger
        self.benefit_service = benefit_service
        self.order_service = order_service

    def execute(self, pos_id):
        # type: (int) -> None

        static_uow = self.benefit_service.get_static_unit_of_work(
            pos_id=pos_id,
            fill_table_info=True,
        )
        static_uow.pos_ot.blkopnotify = True

        updated_order_picture = self.benefit_service.apply_automatic_benefits_on_order(
            pos_id=pos_id,
            static_uow=static_uow,
            on_total=True,
        )

        order_picture = updated_order_picture if updated_order_picture is not None else static_uow.order_picture
        if order_picture.get("state") == "TOTALED":
            product_groups = self.benefit_service.product_groups
            order_picture = self.benefit_service.validate_benefits(
                pos_id=pos_id,
                static_uow=static_uow,
                product_groups=product_groups,
            )
            static_uow.set_order_picture(order_picture=order_picture)

            order_picture = self.benefit_service.remove_invalid_benefits(
                pos_id=pos_id,
                static_uow=static_uow,
            )
            static_uow.set_order_picture(order_picture=order_picture)

            self.benefit_service.apply_pending_benefits(
                pos_id=pos_id,
                static_uow=static_uow,
            )
            self.benefit_service.apply_persisted_benefits(
                pos_id=pos_id,
                static_uow=static_uow,
            )



from typing import Optional


class ApplyContext(object):
    def __init__(
        self,
        pos_id,             # type: int
        originator_pos_id,  # type: int
        user_auth_level,    # type: int
        applier,            # type: Optional[str]
    ):
        self.pos_id = pos_id
        self.originator_pos_id = originator_pos_id
        self.user_auth_level = user_auth_level
        self.applier = applier



from pyscriptscache import ModifierSets, Modifier, Part, cache as _cache


def get_modifiers_and_elements(modifier_sets, modifier_elem, modifier_part, context, part_code, level, pod_type):
    # type: (dict, dict, dict, str, str, int, str) -> None

    # Options - Pai tem default_qty = 1 e Filhos default_qty = 0
    # Ingredients - Pai tem default_qty = 0 e Filhos default_qty = 1
    all_options_tags = [_cache.get_tags_as_dict(part_code, "HasOptions"), _cache.get_tags_as_dict(part_code, "Ingredients")]

    options = []
    for options_tag in all_options_tags:
        if options_tag:
            options.extend(x.split(">")[0] for x in options_tag.split("|"))

    defaults = {}
    options_tag = all_options_tags[1]
    if len(options_tag.split(">")) > 1 and options_tag.split(">")[1]:
        defaults[options_tag.split(">")[0]] = [x for x in options_tag.split(">")[1].split(";")]

    # Busca os valores Default, Min e Max de cada Option ou Ingredients
    qty_option = {x.split(">")[0]: x.split(">")[1].split(";") for x in _cache.get_tags_as_dict(part_code, "QtyOptions").split("|") if x != ""}

    # Busca quais options devem estar apenas no OrderTaker
    dt_only_list = _cache.get_tags_as_dict(part_code, "DTOnly").split("|")
    dt_only_dict = {}
    for dt_only in dt_only_list:
        dt_only_dict[dt_only] = dt_only

    if options:
        temp_modifiers_sets = ModifierSets()
        temp_modifiers_sets.context = context + '.' + part_code
        temp_modifiers_sets.level = level

        has_option = False
        for opt in options:
            if pod_type != "DT" and opt in dt_only_dict:
                continue
            has_option = True

            # OPTION - Code
            item_id = context + '.' + part_code
            temp_mod_set = _cache.get_options(opt)

            if opt in qty_option:
                temp_mod_set.default_qty = qty_option[opt][0]
                temp_mod_set.min_qty = qty_option[opt][1]
                temp_mod_set.max_qty = qty_option[opt][2]
            temp_modifiers_sets.modifiers_sets[opt] = temp_mod_set

            for mod_elem in temp_mod_set.modifiers:  # type: Modifier
                temp_elem = temp_mod_set.modifiers[mod_elem]
                temp_elem.context = item_id + "." + temp_mod_set.part_code

                # Set Default Qty for Ingredients
                if opt in defaults:
                    if mod_elem in defaults[opt]:
                        temp_elem.def_qty = 1
                        temp_elem.ord_qty = 1

                modifier_elem[temp_elem.context + "." + temp_elem.part_code] = temp_elem
                temp_elem.unit_price, temp_elem.add_price, temp_elem.sub_price = _cache.get_best_price(temp_elem.context, temp_elem.part_code, pod_type) or (0, 0, 0)

                # Go to the NEXT LEVEL AND BEYOND
                get_modifiers_and_elements(modifier_sets, modifier_elem, modifier_part, temp_elem.context, temp_elem.part_code, level + 1, pod_type)

        if has_option:
            modifier_sets[temp_modifiers_sets.context] = temp_modifiers_sets

    parts = _cache.get_parts(part_code)
    if parts:
        item_id = context + '.' + part_code
        modifier_part[item_id] = {}

        for part in parts:
            mod_part = Part()
            mod_part.unit_price, mod_part.add_price, mod_part.sub_price = _cache.get_best_price(item_id, part, pod_type) or (0, 0, 0)
            mod_part.part_code = part
            mod_part.context = item_id
            mod_part.def_qty, mod_part.min_qty, mod_part.max_qty, mod_part.product_name = parts[part]

            modifier_part[item_id][part] = mod_part

            # Go to the NEXT LEVEL
            get_modifiers_and_elements(modifier_sets, modifier_elem, modifier_part, item_id, part, level + 1, pod_type)
# END of insert_options_and_defaults



# defusedxml
#
# Copyright (c) 2013 by Christian Heimes <christian@python.org>
# Licensed to PSF under a Contributor Agreement.
# See https://www.python.org/psf/license for licensing details.
"""Defused xml.etree.ElementTree facade
"""
from __future__ import print_function, absolute_import

import sys
import warnings
from xml.etree.ElementTree import TreeBuilder as _TreeBuilder
from xml.etree.ElementTree import parse as _parse
from xml.etree.ElementTree import tostring

from .common import PY3

if PY3:
    import importlib
else:
    from xml.etree.ElementTree import XMLParser as _XMLParser
    from xml.etree.ElementTree import iterparse as _iterparse
    from xml.etree.ElementTree import ParseError


from .common import (
    DTDForbidden,
    EntitiesForbidden,
    ExternalReferenceForbidden,
    _generate_etree_functions,
)

__origin__ = "xml.etree.ElementTree"


def _get_py3_cls():
    """Python 3.3 hides the pure Python code but defusedxml requires it.

    The code is based on test.support.import_fresh_module().
    """
    pymodname = "xml.etree.ElementTree"
    cmodname = "_elementtree"

    pymod = sys.modules.pop(pymodname, None)
    cmod = sys.modules.pop(cmodname, None)

    sys.modules[cmodname] = None
    pure_pymod = importlib.import_module(pymodname)
    if cmod is not None:
        sys.modules[cmodname] = cmod
    else:
        sys.modules.pop(cmodname)
    sys.modules[pymodname] = pymod

    _XMLParser = pure_pymod.XMLParser
    _iterparse = pure_pymod.iterparse
    ParseError = pure_pymod.ParseError

    return _XMLParser, _iterparse, ParseError


if PY3:
    _XMLParser, _iterparse, ParseError = _get_py3_cls()

_sentinel = object()


class DefusedXMLParser(_XMLParser):
    def __init__(
        self,
        html=_sentinel,
        target=None,
        encoding=None,
        forbid_dtd=False,
        forbid_entities=True,
        forbid_external=True,
    ):
        # Python 2.x old style class
        _XMLParser.__init__(self, target=target, encoding=encoding)
        if html is not _sentinel:
            # the 'html' argument has been deprecated and ignored in all
            # supported versions of Python. Python 3.8 finally removed it.
            if html:
                raise TypeError("'html=True' is no longer supported.")
            else:
                warnings.warn(
                    "'html' keyword argument is no longer supported. Pass "
                    "in arguments as keyword arguments.",
                    category=DeprecationWarning,
                )

        self.forbid_dtd = forbid_dtd
        self.forbid_entities = forbid_entities
        self.forbid_external = forbid_external
        if PY3:
            parser = self.parser
        else:
            parser = self._parser
        if self.forbid_dtd:
            parser.StartDoctypeDeclHandler = self.defused_start_doctype_decl
        if self.forbid_entities:
            parser.EntityDeclHandler = self.defused_entity_decl
            parser.UnparsedEntityDeclHandler = self.defused_unparsed_entity_decl
        if self.forbid_external:
            parser.ExternalEntityRefHandler = self.defused_external_entity_ref_handler

    def defused_start_doctype_decl(self, name, sysid, pubid, has_internal_subset):
        raise DTDForbidden(name, sysid, pubid)

    def defused_entity_decl(
        self, name, is_parameter_entity, value, base, sysid, pubid, notation_name
    ):
        raise EntitiesForbidden(name, value, base, sysid, pubid, notation_name)

    def defused_unparsed_entity_decl(self, name, base, sysid, pubid, notation_name):
        # expat 1.2
        raise EntitiesForbidden(name, None, base, sysid, pubid, notation_name)  # pragma: no cover

    def defused_external_entity_ref_handler(self, context, base, sysid, pubid):
        raise ExternalReferenceForbidden(context, base, sysid, pubid)


# aliases
# XMLParse is a typo, keep it for backwards compatibility
XMLTreeBuilder = XMLParse = XMLParser = DefusedXMLParser

parse, iterparse, fromstring = _generate_etree_functions(
    DefusedXMLParser, _TreeBuilder, _parse, _iterparse
)
XML = fromstring


__all__ = [
    "ParseError",
    "XML",
    "XMLParse",
    "XMLParser",
    "XMLTreeBuilder",
    "fromstring",
    "iterparse",
    "parse",
    "tostring",
]



import json
import time
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
from xml.etree import cElementTree as eTree

from helper import round_half_away_from_zero

from application.model.benefit import (
    ActionResult,
    ActionTypes,
    Benefit,
    OrderBenefitInfo,
)
from application.model.exceptions import DiscountNotFound


def benefit_is_add_item_type_on_total(
    benefit,    # type: Dict
):
    # type: (...) -> bool

    actions = benefit.get("actions", {})
    return any([
        action.get("itemsToAdd", False) and action.get("automatic") == "ON_TOTAL"
        for _, action in actions.items()
    ])


def split_benefit_key(
    benefit_key,    # type: str
):
    # type: (...) -> Tuple[str, str]

    benefit_id, applied_time = benefit_key.split("_")
    return benefit_id, applied_time


def get_benefit_to_apply_from_persisted_benefits(
    persisted_benefits,     # type: Dict
    benefit_key=None,       # type: Optional[str]
    default_benefit=None,   # type: Optional[Dict]
):
    # type: (...) -> Dict

    if not benefit_key or not persisted_benefits:
        return default_benefit

    return persisted_benefits["pending_benefits"].get(benefit_key, default_benefit)


def get_benefit_key_from_persisted_benefits(
    benefit_id,             # type: str
    persisted_benefits,     # type: Dict
):
    # type: (...) -> Optional[bool]

    if not benefit_id or not persisted_benefits:
        return None

    persisted_benefits_ids = persisted_benefits.get("pending_benefits_ids", [])

    benefit_key_tuple = filter(lambda b: b[1] == benefit_id, persisted_benefits_ids)

    return "{}_{}".format(benefit_key_tuple[0][1], benefit_key_tuple[0][0]) if benefit_key_tuple else None


def get_benefit_key(
    benefit_id,     # type: str
):
    # type: (...) -> str

    return "{}_{}".format(benefit_id, int(round(time.time() * 1000)))


def get_new_benefit(
    uow,            # type: Any
    action_result,  # type: Optional[ActionResult]
    old_benefits,   # type: List
    benefit_key,    # type: str
    order_state,    # type: str
):
    # type: (...) -> str

    benefit_total_value = sum_discounts(
        promotion_tender=uow.promotion_tenders,
        discount_item_values=action_result.discount_item_values,
    )
    new_benefits = [dict(
        benefit=uow.raw_benefit,
        benefit_id=uow.raw_benefit.get("name"),
        benefit_total_value=benefit_total_value,
        burnt_items=action_result.items_to_burn,
        added_sale_lines=uow.added_sale_lines,
        promotion_tender=uow.promotion_tenders,
        benefit_applier=uow.benefit_applier,
        used_discount_ids=action_result.discounts_id,
        order_tenders_ids=action_result.order_tenders,
        discount_item_values=action_result.discount_item_values,
        benefit_applied=action_result.benefit_applied,
        order_state=order_state,
        benefit_key=benefit_key,
        times_to_apply=uow.benefit.times_to_apply)
    ]

    return json.dumps(old_benefits + new_benefits)


def get_applied_benefits_by_key(
    benefit_keys,       # type: List[str]
    applied_benefits,   # type: List[Dict]
):
    # type: (...) -> List

    matching_benefits = []
    for applied_benefit in applied_benefits:
        applied_benefit_key = applied_benefit.get("benefit_key")
        if applied_benefit_key in benefit_keys:
            matching_benefits.append(applied_benefit)

    if not matching_benefits:
        raise DiscountNotFound()

    return matching_benefits


def is_order_taker(
    model,  # type: eTree
):
    # type: (...) -> bool

    working_mode = model.find(".//WorkingMode")
    pos_function = working_mode.get("posFunction")

    if pos_function is not None and pos_function == "OT":
        return True
    return False


def get_error_message(
    order_id,   # type: int
    benefit,    # type: Benefit
):
    # type: (...) -> str

    return "Error applying benefit for order {} - Benefit name: {}".format(str(order_id), benefit.name)


def get_dict_data_remove(
    benefits,       # type: Dict
    dict_data,      # type: Dict
    property_name,  # type: str
):
    # type: (...) -> Dict

    ids = benefits.get(property_name, {})
    for data_id in dict_data.keys():
        if data_id not in ids.keys():
            ids[data_id] = dict_data[data_id]

    return ids


def get_list_data_remove(
    benefits,       # type: Dict
    list_data,      # type: List
    property_name,  # type: str
):
    # type: (...) -> List

    if len(benefits) == 0:
        return []

    ids = benefits.get(property_name, "[]")
    for data_id in list_data:
        if data_id not in ids:
            ids.append(data_id)

    return ids


def remove_dict_properties_benefits(
    benefits,       # type: Dict
    dict_remove,    # type: Dict
    property_name,  # type: str
):
    # type: (...) -> Dict

    result = {}
    body = benefits.get(property_name, {})
    for key in body.keys():
        if key not in dict_remove.keys():
            result[key] = body[key]

    return result


def remove_list_properties_benefits(
    benefits,       # type: Dict
    list_remove,    # type: List
    property_name,  # type: str
):
    # type: (...) -> List

    result = []
    keys = benefits.get(property_name, "[]")
    for key in keys:
        if key not in list_remove:
            result.append(key)

    return result


def sum_discount_items(
    discount_item_values,   # type: Dict
):
    # type: (...) -> float

    total_discount = 0
    for key, value in discount_item_values.items():
        if isinstance(value, float):
            total_discount += value
            continue

        total_discount += sum_discount_items(discount_item_values=value)

    return total_discount


def sum_discounts(
    promotion_tender,       # type: Dict
    discount_item_values,   # type: Dict
):
    # type: (Dict, Dict) -> float

    sum_promotion_tender = sum([float(x.get("amount")) for x in promotion_tender])
    sum_discount_item_values = sum_discount_items(discount_item_values=discount_item_values)

    return sum_promotion_tender + sum_discount_item_values


def get_remaining_amount_discount(
    max_value_to_discount,  # type: float
    actual_value=0,         # type: float
):
    # type: (...) -> float

    if max_value_to_discount is None:
        return float("inf")

    return max_value_to_discount - actual_value


def get_line_id_from_sale_line(
    sale_line,          # type: Any
    from_xml=False,     # type: Optional[bool]
):
    # type: (...) -> str

    if from_xml:
        return "{}.{}-{}.{}".format(
            sale_line.get("lineNumber"),
            sale_line.get("level"),
            sale_line.get("itemId"),
            sale_line.get("partCode"),
        )

    return "{}.{}-{}.{}".format(
        sale_line.line_number,
        sale_line.level,
        sale_line.context,
        sale_line.item_id,
    )


def round_value(
    value,      # type: float
    decimal=2,  # type: Optional[int]
):
    # type: (...) -> Optional[Union[float, int]]

    if value is not None:
        return round_half_away_from_zero(
            number=value,
            precision=decimal,
        )

    return None


def is_invalid_sale_line_or_remaining_quantity(
    restriction,            # type: Restrictions
    sale_line,              # type: SaleLine
    remaining_quantity,     # type: Any[Union[int, float]]
):
    # type: (...) -> bool

    return not restriction.validate(sale_line) or (remaining_quantity is not None and remaining_quantity == 0)


def subtract_quantity(
    value,              # type: Any[Union[int, float]]
    value_to_subtract,  # type: Any[Union[int, float]]
):
    # type: (...) -> Any[Union[int, float]]

    return None if value is None else value - value_to_subtract


def add_to_dict_value(
    object_dict,    # type: Dict
    dict_key,       # type: Any
    value,          # type: Any[Union[int, float]]
):
    # type: (...) -> None

    if dict_key in object_dict.keys():
        object_dict[dict_key] += value
    else:
        object_dict[dict_key] = value


def have_action_by_items_or_groups(
    benefit,            # type: Benefit
    on_total=False,     # type: Optional[bool]
):
    # type: (...) -> bool

    if on_total:
        action_types = [
            ActionTypes.discountAmountByItemsOnTotal.value,
            ActionTypes.discountPercentageByItemsOnTotal.value,
            ActionTypes.discountAmountByGroupsOnTotal,
            ActionTypes.discountPercentageByGroupsOnTotal,
        ]
    else:
        action_types = [
            ActionTypes.discountAmountByItems.value,
            ActionTypes.discountPercentageByItems.value,
        ]

    for action in benefit.actions:
        if action.name in action_types:
            return True

    return False


def get_item_values_line_ids(
    benefit,    # type: OrderBenefitInfo
):
    item_values = benefit.discount_item_values.values()
    item_values_line_ids = []
    for item_value in item_values:
        item_values_line_ids.extend(item_value.keys())
    return item_values_line_ids



from modelcommon.printer import SeparatorTag

import constants
from report.command import CutPaperCommand

from .._CommandProcessor import CommandProcessor


class CutPaperCommandProcessor(CommandProcessor):
    def get_bytes(self, command):
        if not isinstance(command, CutPaperCommand):
            return ""

        return b'{0}SpaceCutByTable{1}'.format(SeparatorTag.start, SeparatorTag.end)



from application.repository.pos.benefitapplier.loyalty._LoyaltyCouponApplierRepository import LoyaltyCouponApplierRepository
from ._NDiscountCouponApplierRepository import NDiscountCouponApplierRepository



import json
from abc import abstractmethod
from typing import Any

from messagebus import (
    DataType,
    Message,
    MessageBus,
    AckMessage,
    InternalErrorMessage,
)
from messageprocessorutil import UuidMessageProcessor
from .dto import ApplyCouponErrorCode
from ..dto import ResponseToken
from ...model.customexception import ErrorApplyingBenefit
from ...model.repository.benefit_applier.exception import (
    CouponAlreadyUsed,
    CouponLocked,
    CouponCustomError,
    BenefitApplierNotFound,
)


class BaseApplyProcessor(UuidMessageProcessor):
    @abstractmethod
    def parse_data(
        self,        # type: BaseApplyProcessor
        data,        # type: bytes
        event_type,  # type: str
    ):
        # type: (...) -> None

        raise NotImplementedError()

    @abstractmethod
    def call_business(
            self,     # type: BaseApplyProcessor
            request,  # type: Any
    ):
        # type: (...) -> None

        raise NotImplementedError()

    def format_response(
        self,         # type: BaseApplyProcessor
        message_bus,  # type: MessageBus
        message,      # type: Message
        event,        # type: None
        request,      # type: Any
        response,     # type: None
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=AckMessage()
        )

    def format_exception(
        self,         # type: BaseApplyProcessor
        message_bus,  # type: MessageBus
        message,      # type: Message
        event,        # type: None
        input_obj,    # type: Any
        exception,    # type: Exception
    ):
        # type: (...) -> None

        exception_map = {
            CouponAlreadyUsed: (ApplyCouponErrorCode.COUPON_ALREADY_USED, None),
            CouponLocked: (ApplyCouponErrorCode.COUPON_LOCKED, None),
            CouponCustomError: (ApplyCouponErrorCode.CUSTOM_ERROR, self._build_data_from_coupon_custom_error),
            BenefitApplierNotFound: (ApplyCouponErrorCode.COUPON_NOT_FOUND, None),
            ErrorApplyingBenefit: (ApplyCouponErrorCode.NDISCOUNT_ERROR, self._build_data_from_ndiscount_error)
        }
        error_code, data_builder = exception_map.get(type(exception), (None, None))

        if error_code is not None:
            data = error_code.name
            if data_builder is not None:
                data += "\0" + data_builder(exception)

            message_bus.reply_message(
                message=message,
                reply_message=Message(
                    token=ResponseToken.TK_NCOUPON_ERROR,
                    data_type=DataType.param,
                    data=data,
                )
            )
        else:
            message_bus.reply_message(
                message=message,
                reply_message=InternalErrorMessage(exception),
            )

    def format_parse_exception(
        self,         # type: BaseApplyProcessor
        message_bus,  # type: MessageBus
        message,      # type: Message
        event,        # type: None
        data,         # type: bytes
        exception,    # type: Exception
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=InternalErrorMessage(exception),
        )

    @staticmethod
    def _build_data_from_coupon_custom_error(
        exception,  # type: CouponCustomError
    ):
        # type: (...) -> bytes

        return json.dumps({
            "message": {
                "pt": exception.message.pt,
                "en": exception.message.en,
                "es": exception.message.es,
            }
        }).encode("utf-8")

    @staticmethod
    def _build_data_from_ndiscount_error(
        exception,  # type: ErrorApplyingBenefit
    ):
        # type: (...) -> str

        return str(exception)



import json
from logging import Logger
from xml.etree import cElementTree as eTree

from model.config import Config
from model.repository import PosRepositoryModel
from model.services import (
    CouponServiceModel,
    OrderServiceModel
)
from model.error_type import PosValidationErrorsType
from model.exception import (
    ValidationError,
    ExternalServiceErrors,
)
from validation_util import (
    check_if_pos_id_is_valid,
    validate_api_key,
    check_if_order_is_in_progress_or_totaled,
)
from posot import OrderTaker

from interactor.route.order.get_coupons.dto import GetCouponRequest

from typing import (
    List,
    Optional,
)


class GetCouponInteractor(object):
    def __init__(
        self,
        logger,             # type: Logger
        config,             # type: Config
        pos_repository,     # type: PosRepositoryModel
        order_service       # type: OrderServiceModel
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.order_service = order_service
        self.model = None
        self.pos_ot = None  # type: Optional[OrderTaker]
        self.price_list = None

    def execute(self, request):
        # type: (GetCouponRequest) -> List

        pos_id = validate_api_key(
            request=request,
            secure_api_key_list=self.config.api_key_mapping,
        )
        check_if_pos_id_is_valid(
            repository=self.pos_repository,
            pos_id=pos_id,
        )

        self.model = self.pos_repository.get_model(pos_id=pos_id)
        self.pos_ot = self.pos_repository.get_pos_ot(model=self.model)

        if self.pos_repository.has_current_order(model=self.model):
            check_if_order_is_in_progress_or_totaled(
                repository=self.pos_repository,
                model=self.model
            )

        try:
            order = self.pos_repository.get_current_order(model=self.model)
            order_id = order.get('orderId')
            order = self.pos_repository.get_order_details(
                pos_id=pos_id,
                order_id=order_id,
                pos_ot=self.pos_ot
            )
        except Exception as ex:
            error_message = "Error retrieving applied coupons. Pos ID: {}".format(pos_id)
            self.logger.error(
                msg=error_message,
                exc_info=ex,
            )
            raise ExternalServiceErrors(PosValidationErrorsType.CouponValidationError, error_message)

        if order:
            return self._get_applied_coupons_details(order=order)

        error_message = "Error retrieving applied coupons. Order: {}".format(order_id)
        self.logger.error(error_message)
        raise ValidationError(PosValidationErrorsType.CouponValidationError, error_message)

    def _get_applied_coupons_details(
        self,
        order,      # type: eTree
    ):
        # type: (...) -> List

        benefits = order.find(".//OrderProperty[@key=\"BENEFIT_LIST\"]")
        if benefits is not None:
            benefits = json.loads(benefits.get("value"))
            benefits = json.loads(benefits.get("benefitList"))
            benefits_details = []
            for benefit in benefits:
                benefit_dict = {
                    "name": benefit.get("benefit").get("userDescription"),
                    "id": benefit.get("benefit_key"),
                    "value": benefit.get("benefit_total_value"),
                }
                benefits_details.append(benefit_dict)

            self.logger.info("Success retrieving applied coupons. Coupons: {}".format(benefits))
            return benefits_details

        return []



from old_helper import BaseRepository
from helper import remove_accents

tender_types_cash = {}
grouped_tender_types = {}


class ProductRepository(BaseRepository):
    def __init__(self, mbcontext):
        super(ProductRepository, self).__init__(mbcontext)

    def populate_card_brands_groups(self, card_brands_payments):
        if len(tender_types_cash) == 0:
            self._get_tender_types()
            self._group_card_brands()

        populated_brands = {}
        for tender_id in card_brands_payments:
            tender_name = self.get_tender_name(tender_id).title()

            card_brand = {
                tender_id: {
                    'value': card_brands_payments[tender_id]['value'],
                    'name': tender_name,
                    'quantity': card_brands_payments[tender_id]['quantity'],
                    'id': tender_id
                }
            }

            if tender_id in grouped_tender_types:
                if tender_id in populated_brands:
                    populated_brands[tender_id].update(card_brand)
                else:
                    populated_brands[tender_id] = card_brand
            else:
                parent_id = self.get_parent_id(tender_id)
                if parent_id in populated_brands:
                    parent_brands = populated_brands[parent_id]
                    if tender_id in parent_brands:
                        parent_brands[tender_id]['value'] += card_brands_payments[tender_id]['value']
                        parent_brands[tender_id]['quantity'] += card_brands_payments[tender_id]['quantity']
                    else:
                        parent_brands.update(card_brand)
                else:
                    populated_brands[parent_id] = card_brand

        from collections import OrderedDict
        for tender in populated_brands:
            populated_brands[tender] = OrderedDict(sorted(populated_brands[tender].items()))
        return OrderedDict(sorted(populated_brands.items()))

    @staticmethod
    def get_tender_name(tender_id):
        if tender_id in tender_types_cash:
            return remove_accents(tender_types_cash[tender_id][0])
        return "OUTROS"

    @staticmethod
    def get_parent_id(tender_id):
        for parent_id in grouped_tender_types:
            if tender_id in grouped_tender_types[parent_id]:
                return parent_id

    @staticmethod
    def _group_card_brands():
        for tender_id in tender_types_cash:
            tender_parent_id = tender_types_cash[tender_id][1]
            if tender_parent_id is not None:
                if int(tender_parent_id) in grouped_tender_types:
                    grouped_tender_types[int(tender_parent_id)].append(tender_id)
                else:
                    grouped_tender_types[int(tender_parent_id)] = [tender_id]
            else:
                if tender_id not in grouped_tender_types:
                    grouped_tender_types[tender_id] = []

    def _get_tender_types(self):
        def inner_func(conn):
            query = _GetTenderTypes
            cursor = conn.select(query)
            for row in cursor:
                tender_types_cash[int(row.get_entry(0))] = (str(row.get_entry(1)), row.get_entry(2))

        return self.execute_with_connection(inner_func)


_GetTenderTypes = "SELECT TenderId, TenderDescr, ParentTenderId FROM TenderType"



# Copyright 2009-2017 Wander Lairson Costa
# Copyright 2009-2020 PyUSB contributors
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

r"""usb.control - USB standard control requests

This module exports:

get_status - get recipeint status
clear_feature - clear a recipient feature
set_feature - set a recipient feature
get_descriptor - get a device descriptor
set_descriptor - set a device descriptor
get_configuration - get a device configuration
set_configuration - set a device configuration
get_interface - get a device interface
set_interface - set a device interface
"""

__author__ = 'Wander Lairson Costa'

__all__ = ['get_status',
           'clear_feature',
           'set_feature',
           'get_descriptor',
           'set_descriptor',
           'get_configuration',
           'set_configuration',
           'get_interface',
           'set_interface',
           'ENDPOINT_HALT',
           'FUNCTION_SUSPEND',
           'DEVICE_REMOTE_WAKEUP',
           'U1_ENABLE',
           'U2_ENABLE',
           'LTM_ENABLE']

import usb.util as util
import usb.core as core

def _parse_recipient(recipient, direction):
    if recipient is None:
        r = util.CTRL_RECIPIENT_DEVICE
        wIndex = 0
    elif isinstance(recipient, core.Interface):
        r = util.CTRL_RECIPIENT_INTERFACE
        wIndex = recipient.bInterfaceNumber
    elif isinstance(recipient, core.Endpoint):
        r = util.CTRL_RECIPIENT_ENDPOINT
        wIndex = recipient.bEndpointAddress
    else:
        raise ValueError('Invalid recipient.')
    bmRequestType = util.build_request_type(
                            direction,
                            util.CTRL_TYPE_STANDARD,
                            r
                        )
    return (bmRequestType, wIndex)

# standard feature selectors from USB 2.0/3.0
ENDPOINT_HALT = 0
FUNCTION_SUSPEND = 0
DEVICE_REMOTE_WAKEUP = 1
U1_ENABLE = 48
U2_ENABLE = 49
LTM_ENABLE = 50

def get_status(dev, recipient = None):
    r"""Return the status for the specified recipient.

    dev is the Device object to which the request will be
    sent to.

    The recipient can be None (on which the status will be queried
    from the device), an Interface or Endpoint descriptors.

    The status value is returned as an integer with the lower
    word being the two bytes status value.
    """
    bmRequestType, wIndex = _parse_recipient(recipient, util.CTRL_IN)
    ret = dev.ctrl_transfer(bmRequestType = bmRequestType,
                            bRequest = 0x00,
                            wIndex = wIndex,
                            data_or_wLength = 2)
    return ret[0] | (ret[1] << 8)

def clear_feature(dev, feature, recipient = None):
    r"""Clear/disable a specific feature.

    dev is the Device object to which the request will be
    sent to.

    feature is the feature you want to disable.

    The recipient can be None (on which the status will be queried
    from the device), an Interface or Endpoint descriptors.
    """
    if feature == ENDPOINT_HALT:
        dev.clear_halt(recipient)
    else:
        bmRequestType, wIndex = _parse_recipient(recipient, util.CTRL_OUT)
        dev.ctrl_transfer(bmRequestType = bmRequestType,
                          bRequest = 0x01,
                          wIndex = wIndex,
                          wValue = feature)

def set_feature(dev, feature, recipient = None):
    r"""Set/enable a specific feature.

    dev is the Device object to which the request will be
    sent to.

    feature is the feature you want to enable.

    The recipient can be None (on which the status will be queried
    from the device), an Interface or Endpoint descriptors.
    """
    bmRequestType, wIndex = _parse_recipient(recipient, util.CTRL_OUT)
    dev.ctrl_transfer(bmRequestType = bmRequestType,
                      bRequest = 0x03,
                      wIndex = wIndex,
                      wValue = feature)

def get_descriptor(dev, desc_size, desc_type, desc_index, wIndex = 0):
    r"""Return the specified descriptor.

    dev is the Device object to which the request will be
    sent to.

    desc_size is the descriptor size.

    desc_type and desc_index are the descriptor type and index,
    respectively. wIndex index is used for string descriptors
    and represents the Language ID. For other types of descriptors,
    it is zero.
    """
    wValue = desc_index | (desc_type << 8)

    bmRequestType = util.build_request_type(
                        util.CTRL_IN,
                        util.CTRL_TYPE_STANDARD,
                        util.CTRL_RECIPIENT_DEVICE)

    desc = dev.ctrl_transfer(
            bmRequestType = bmRequestType,
            bRequest = 0x06,
            wValue = wValue,
            wIndex = wIndex,
            data_or_wLength = desc_size)

    if len(desc) < 2:
        raise USBError('Invalid descriptor')

    return desc

def set_descriptor(dev, desc, desc_type, desc_index, wIndex = None):
    r"""Update an existing descriptor or add a new one.

    dev is the Device object to which the request will be
    sent to.

    The desc parameter is the descriptor to be sent to the device.
    desc_type and desc_index are the descriptor type and index,
    respectively. wIndex index is used for string descriptors
    and represents the Language ID. For other types of descriptors,
    it is zero.
    """
    wValue = desc_index | (desc_type << 8)

    bmRequestType = util.build_request_type(
                        util.CTRL_OUT,
                        util.CTRL_TYPE_STANDARD,
                        util.CTRL_RECIPIENT_DEVICE)

    dev.ctrl_transfer(
        bmRequestType = bmRequestType,
        bRequest = 0x07,
        wValue = wValue,
        wIndex = wIndex,
        data_or_wLength = desc)

def get_configuration(dev):
    r"""Get the current active configuration of the device.

    dev is the Device object to which the request will be
    sent to.

    This function differs from the Device.get_active_configuration
    method because the later may use cached data, while this
    function always does a device request.
    """
    bmRequestType = util.build_request_type(
                            util.CTRL_IN,
                            util.CTRL_TYPE_STANDARD,
                            util.CTRL_RECIPIENT_DEVICE)

    return dev.ctrl_transfer(
                bmRequestType,
                bRequest = 0x08,
                data_or_wLength = 1)[0]

def set_configuration(dev, bConfigurationNumber):
    r"""Set the current device configuration.

    dev is the Device object to which the request will be
    sent to.
    """
    dev.set_configuration(bConfigurationNumber)

def get_interface(dev, bInterfaceNumber):
    r"""Get the current alternate setting of the interface.

    dev is the Device object to which the request will be
    sent to.
    """
    bmRequestType = util.build_request_type(
                            util.CTRL_IN,
                            util.CTRL_TYPE_STANDARD,
                            util.CTRL_RECIPIENT_INTERFACE)

    return dev.ctrl_transfer(
                bmRequestType = bmRequestType,
                bRequest = 0x0a,
                wIndex = bInterfaceNumber,
                data_or_wLength = 1)[0]

def set_interface(dev, bInterfaceNumber, bAlternateSetting):
    r"""Set the alternate setting of the interface.

    dev is the Device object to which the request will be
    sent to.
    """
    dev.set_interface_altsetting(bInterfaceNumber, bAlternateSetting)




from abc import ABCMeta, abstractmethod
from typing import Optional

from ._FoundRoute import FoundRoute
from ._Route import Route


class HttpRouter(object):
    __meta__ = ABCMeta

    @abstractmethod
    def get_route(self, route):
        # type: (Route) -> Optional[FoundRoute]
        raise NotImplementedError()



from _DefaultHeaderFormatter import DefaultHeaderFormatter
from basereport._TableReport import TableReport
from _Translator import Translator
from _EscPosImage import EscPosImage
from _EscPosQrGenerator import EscPosQrCodeGenerator
from _QrCodeGenerator import QrCodeGenerator
from _LogoImage import LogoImage

__all__ = ["EscPosImage",
           "EscPosQrCodeGenerator",
           "QrCodeGenerator",
           "Translator",
           "TableReport",
           "DefaultHeaderFormatter",
           "LogoImage"]



from typing import Dict


class DictionaryOperations(object):
    @staticmethod
    def merge_counting_dicts(items_list):
        # type: (list[dict]) -> dict

        response = {}
        for d in items_list:
            for key, value in d.items():
                response[key] = value + response.get(key, 0)

        return response

    @staticmethod
    def unmerge_counting_dict(subtract_from, items):
        # type: (Dict, Dict) -> Dict

        response = subtract_from.copy()
        for key, value in items.items():
            if response.get(key) is not None:
                response[key] = response.get(key, 0) - value

        return response



import os
import sqlite3

from sqlite3 import Connection
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Tuple,
    Union
)

from dbapitransactioncallback import DbApiTransactionCallback
from dbapitemplate import DbApiTemplate
from repository import FiscalRepository as BaseFiscalRepository


class SqliteFiscalRepository(BaseFiscalRepository):

    def __init__(
        self,
        database_dir,           # type: str
        backup_database_dir,    # type: str
        schema_sql_path,        # type: str
        callback                # type: DbApiTransactionCallback
    ):
        # type: (...) -> None

        self.callback = callback
        self.database_path = os.path.join(database_dir, "fiscal_persistcomp.db")
        self.backup_database_dir = backup_database_dir
        self.schema_sql_path = schema_sql_path

        self.fiscal_conn = None
        self.template = None
        self.sale_types = None

    def _get_db_api_template(
            self
    ):
        # type: () -> None

        if self.template is not None:
            return

        fiscal_conn = self._get_fiscal_conn()
        self.template = DbApiTemplate(fiscal_conn)
        self._update_callback_conn(fiscal_conn)

    def _get_fiscal_conn(
        self
    ):
        # type: () -> Connection

        if self.fiscal_conn is None:
            self.fiscal_conn = sqlite3.connect(self.database_path, check_same_thread=False)

        return self.fiscal_conn

    def _update_callback_conn(
        self,
        conn    # type: Connection
    ):
        # type: (...) -> None

        self.callback.update_conn(conn)

    @staticmethod
    def _execute_script(
        connection,     # type: Connection
        query           # type: str
    ):
        # type: (...) -> None

        connection.executescript(query)

    def _execute_template(
        self,
        query,              # type: str
        params=None,        # type: Tuple
        mapper=None,        # type: Optional[Callable[[tuple], Any]]
        is_script=False     # type: Optional[bool]
    ):
        # type: (...) -> Union[int, list]

        self._get_db_api_template()

        if is_script:
            return self._execute_script(
                connection=self.template.conn,
                query=query
            )
        return self.template.execute(
            sql=query,
            params=params if params else (), mapper=mapper
        )

    def get_orders(
        self,
        end_date    # type: int
    ):
        # type: (...) -> List[Tuple[int, str]]

        query = self._get_orders_by_date_query

        params = (end_date,) if end_date else None

        def mapper(row):
            return row["OrderId"], row["DataNota"]

        return self._execute_template(
            query=query,
            params=params,
            mapper=mapper
        )

    def _get_create_tables_query(self):
        # type: () -> str

        try:
            with open(self.schema_sql_path, "r") as f:
                query = f.read()
            return query
        except Exception as _:
            return ""

    def _create_tables_fiscals(self, db_api):
        # type: (DbApiTemplate) -> None

        query = self._get_table_exists_query
        _, data = db_api.execute(query, ())

        if not data:
            create_table_query = self._get_create_tables_query()
            self._execute_script(
                connection=db_api.conn,
                query=create_table_query
            )

    def fill_fiscal_backup_database(
        self,
        suffix_date,    # type: str
        orders          # type: List[int]
    ):
        # type: (...) -> None

        conn = sqlite3.connect(os.path.join(self.backup_database_dir, "fiscal_{}.db".format(suffix_date)))
        self._update_callback_conn(conn)
        db_api = DbApiTemplate(conn)

        self._create_tables_fiscals(db_api)

        query = self._backup_fiscal_database_query.format(self.database_path, ','.join([str(o) for o in orders]))

        self._execute_script(
            connection=db_api.conn,
            query=query
        )

    def delete_fiscals(
        self,
        orders      # type: List[int]
    ):
        # type: (...) -> None

        query = self._get_delete_orders_query.format(','.join([str(o) for o in orders]))
        self._execute_template(
            query=query,
            is_script=True
        )

    _get_table_exists_query = """
        SELECT count(name) FROM sqlite_master WHERE type='table' AND name='FiscalData'
    """

    _get_orders_by_date_query = """
        SELECT OrderId, DataNota
        FROM FiscalData
        WHERE SentToNfce = 1 AND ? > DataNota
        ORDER BY OrderId LIMIT 1000
    """

    _backup_fiscal_database_query = """
        ATTACH DATABASE '{0}' AS f; 
        INSERT OR REPLACE INTO FiscalData SELECT * FROM f.FiscalData WHERE OrderId in ({1});
        INSERT OR REPLACE INTO PaymentData SELECT * FROM f.PaymentData WHERE OrderId in ({1});
        DETACH f;
    """

    _get_delete_orders_query = """
        DELETE FROM FiscalData WHERE OrderId in ({0});
        DELETE FROM PaymentData WHERE OrderId in ({0});
    """



from _ContextKeys import ContextKeys
from _NfceXmlPartBuilder import NfceXmlPartBuilder
from models import SoftwareDeveloper
from pos_model import Order
from typing import Dict, Any


class InfRespTecBuilder(NfceXmlPartBuilder):

    def __init__(
            self,
            software_developer  # type: SoftwareDeveloper
    ):
        # type: (...) -> None

        self.software_developer = software_developer

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        xml = ""

        if self.software_developer.enabled is True:
            xml += "<infRespTec>"

            xml += "<CNPJ>{}</CNPJ>".format(self.software_developer.cnpj)
            xml += "<xContato>{}</xContato>".format(self.software_developer.contact)
            xml += "<email>{}</email>".format(self.software_developer.email)
            xml += "<fone>{}</fone>".format(self.software_developer.phone)

            if self.software_developer.id_csrt and self.software_developer.csrt:
                xml += "<idCSRT>{}</idCSRT>".format(self.software_developer.id_csrt)
                xml += "<hashCSRT>${HASH_CSRT}</hashCSRT>"

                context[ContextKeys.csrt] = self.software_developer.csrt
                context[ContextKeys.id_csrt] = self.software_developer.id_csrt

            xml += "</infRespTec>"

        return xml



import json
from xml.etree import cElementTree as eTree

from messagebus import DefaultToken, DataType, Message
from messageprocessorutil import UuidMessageProcessor

from typing import Optional

from application.interactor import OnSaleItemInteractor
from application.model import Configurations


class OnSaleItemProcessor(UuidMessageProcessor):
    def __init__(self, configs, interactor):
        # type: (Configurations, OnSaleItemInteractor) -> None
        super(OnSaleItemProcessor, self).__init__()
        self.logger = configs.logger
        self.configs = configs
        self.interactor = interactor

    def parse_data(self, data, event_type):
        # type: (str, str) -> (int, int, int, str, eTree)

        data = json.loads(data)
        pos_id = int(data.get("pos_id"))

        return pos_id

    def call_business(self, pos_id):
        # type: (int) -> Optional[eTree.XML]

        return self.interactor.execute(pos_id=pos_id)

    def format_response(self, message_bus, message, event, request, response):
        if message:
            data = eTree.tostring(element=response) if response is not None else ""
            message_bus.reply_message(
                message=message,
                reply_message=Message(
                    token=DefaultToken.TK_SYS_ACK,
                    data_type=DataType.string,
                    data=data,
                ),
            )

    def format_exception(self, message_bus, message, event, input_obj, exception):
        if message:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        if message:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



class ContextKeys(object):
    total_vbc = "total_vbc"
    total_prod = "total_prod"
    total_icms = "total_icms"
    total_pis = "total_pis"
    total_cofins = "total_cofins"
    item_no = "item_no"
    current_sale_item = "current_sale_item"
    sale_items = "sale_items"
    data_emissao = "data_emissao"
    cpf_cnpj = "cpf_cnpj"



import base64
import json
import time
from logging import Logger
from xml.etree import cElementTree as eTree

from model.error_type import (
    ExternalServicesErrorsType,
    OperationErrorsType,
    PosValidationErrorsType,
)
from model.exception import (
    ExternalServiceErrors,
    OperationErrors,
    ValidationError,
)
from interactor.route.payment.insert.dto import OrderInsertPaymentRequest
from model.repository import PosRepositoryModel
from model.services import FiscalServiceModel
from model.tokens import BusTokens
from interactor.route.electronic_payment.send_responses.dto import SendResponsesRequest
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Message, DataType
from msgbus import (
    TK_SYS_ACK,
    TK_SYS_NAK,
)
from mwhelper import BaseRepository
from persistence import Connection
from sysactions import (
    get_used_service,
    translate_message,
)
from typing import (
    Dict,
    List,
    Optional,
)

from services import PrinterService


class FiscalService(FiscalServiceModel, BaseRepository):

    def __init__(
        self,
        message_bus,        # type: MbContextMessageBus
        logger,             # type: Logger
        pos_repository,     # type: PosRepositoryModel
        printer_service,    # type: PrinterService
    ):
        # type: (...) -> None

        super(FiscalService, self).__init__(message_bus.mbcontext)
        self.message_bus = message_bus
        self.logger = logger
        self.electronic_payments_id = None
        self.pos_repository = pos_repository
        self.printer_service = printer_service

    def process_order_fiscal(
        self,
        model,          # type: eTree
        print_coupon,   # type: bool
        device_type,    # type: str
    ):
        # type: (...) -> Optional[str]

        max_retry = 2
        while True:
            fiscal_response = self._fiscalize_order(
                model=model,
                device_type=device_type,
            )

            self.logger.info("FiscalResponse: {}".format(fiscal_response.data))

            if fiscal_response.token == TK_SYS_ACK:
                if print_coupon:
                    self.logger.info("Start print fiscal coupon")
                    self._print_fiscal_coupon(
                        model=model,
                        fiscal_response=fiscal_response,
                    )
                    self.logger.info("End print fiscal coupon")
                return fiscal_response.data

            fiscal_certificate_erros = [
                "Certificado da Loja Expirado",
                "Falha na Leitura da Validade do Certificado NFCE",
            ]
            if fiscal_response.data in fiscal_certificate_erros:
                raise ValidationError(PosValidationErrorsType.FiscalCertificateError)

            message = ""
            if len(fiscal_response.data.split('\0')) < 2:
                is_fiscal_ok = "False"
            elif len(fiscal_response.data.split('\0')) == 2:
                is_fiscal_ok, message = fiscal_response.data.split('\0')
            else:
                is_fiscal_ok, message = fiscal_response.data.split('\0', 1)

            if is_fiscal_ok == "True":
                self.logger.info("Order fiscalized but without fiscal coupon print")
                break

            if max_retry > 0:
                max_retry -= 1
                self.logger.error("Cannot fiscalize coupon. Remaining attempts: {}".format(max_retry))
                time.sleep(1)
                continue

            if is_fiscal_ok == "False" and message.startswith("$"):
                message = message[1:] if message.startswith("$") else message
                params = ""
                if "|" in message:
                    message, params = message.split("|", 1)
                message = translate_message(model, message, *(params.split("|")))
                raise ExternalServiceErrors(ExternalServicesErrorsType.FiscalizeError, message)

            raise StopIteration()

    def _print_fiscal_coupon(
        self,               # type: FiscalService
        model,              # type: eTree
        fiscal_response,    # type: Message
    ):
        # type: (...) -> None

        try:
            print_response = self.printer_service.send_print_data(
                model=model,
                response=fiscal_response.data,
            )
            if print_response is None or print_response.token == TK_SYS_NAK:
                self.logger.error("Cannot print fiscal coupon")
        except Exception as _:
            self.logger.exception("Error printing fiscal coupon")

    def _fiscalize_order(
        self,
        model,          # type: eTree
        device_type,    # type: str
    ):
        # type: (...) -> Message

        order = self._get_current_order(model)
        pos_id = order.get("posId")
        order_id = order.get("orderId")
        data = "\0".join([pos_id, order_id, device_type])

        log_message = "Fiscalize order: [pos_id: {} order_id: {} device_type: {}]"
        self.logger.info(log_message.format(pos_id, order_id, device_type))

        message = Message(token=BusTokens.TK_FISCALWRAPPER_PROCESS_REQUEST, data_type=DataType.param, data=data)
        return self.message_bus.send_message("FiscalWrapper", message)

    def process_payment(
        self,
        model,      # type: eTree
        request     # type: OrderInsertPaymentRequest
    ):
        # type: (...) -> Optional[str]

        log_message = "Process payment: save promotional tenders payment data: [pos_id: {} order_id: {}]"
        self.logger.info(log_message.format(request.pos_id, request.order_id))
        self._save_promotional_tenders_payment_data(model)

        if request.electronic_type_id in self.electronic_payments_id:
            return self._process_electronic_payment(
              model=model,
              request=request,
            )

        return None

    def _process_electronic_payment(
        self,
        model,      # type: eTree
        request     # type: OrderInsertPaymentRequest
    ):
        # type: (...) -> str

        log_message = "Process payment: get electronic payment service: [pos_id: {} order_id: {}]"
        self.logger.info(log_message.format(request.pos_id, request.order_id))
        payment_service = self._get_electronic_payment_service(model)
        order_picture = None
        try:
            log_message = "Process payment: get order picture: [pos_id: {} order_id: {}]"
            self.logger.info(log_message.format(request.pos_id, request.order_id))
            order_picture = self.pos_repository.get_order_picture(
                pos_id=request.pos_id,
                order_id=request.order_id,
                model=model,
            )

            log_message = "Process payment: add electronic payment: [pos_id: {} order_id: {}]"
            self.logger.info(log_message.format(request.pos_id, request.order_id))
            response = self._add_electronic_payment(
                model=model,
                order=order_picture,
                payment_service=payment_service,
                pos_id=str(request.pos_id),
                amount=request.amount,
                electronic_type_id=str(request.electronic_type_id),
            )

            log_message = "Process payment: validate electronic payment response: [pos_id: {} order_id: {}]"
            self.logger.info(log_message.format(request.pos_id, request.order_id))
            request.amount = self._validate_electronic_payment_response(
                response=response,
                amount=request.amount,
                pos_id=str(request.pos_id),
                order_id=str(request.order_id),
            )
        except Exception as _:
            message = "Error processing electronic payment: [pos_id: {} order_id: {}]: \norderPicture: {}"
            self.logger.exception(message.format(request.pos_id, request.order_id, eTree.tostring(order_picture)))
            self._cancel_electronic_payment(payment_service)
            raise

        log_message = "Process payment: get electronic fiscal params: [pos_id: {} order_id: {}]"
        self.logger.info(log_message.format(request.pos_id, request.order_id))
        fiscal_params = self._get_electronic_fiscal_params(
            model=model,
            pos_id=request.pos_id,
            order_id=request.order_id,
            payment_id=request.payment_id,
            amount=request.amount,
            response=response,
        )

        log_message = "Process payment: store payment data: [pos_id: {} order_id: {}]"
        self.logger.info(log_message.format(request.pos_id, request.order_id))
        self._store_payment_data(fiscal_params)

        log_message = "Process payment: get tender details and Finalize: [pos_id: {} order_id: {}]"
        self.logger.info(log_message.format(request.pos_id, request.order_id))
        return self._get_tender_details(eTree.fromstring(response.data))

    def _add_electronic_payment(
        self,
        model,                # type: eTree
        order,                # type: eTree
        payment_service,      # type: str
        pos_id,               # type: str
        amount,               # type: float
        electronic_type_id,   # type: str
    ):
        # type: (...) -> Message

        data = self._get_electronic_payment_data(
            order=order,
            model=model,
            pos_id=pos_id,
            amount=amount,
            electronic_type_id=electronic_type_id,
        )
        message = Message(token=BusTokens.TK_ELECTRONIC_PAYMENT_ADD_PAYMENT, data_type=DataType.param, data=data)
        return self.message_bus.send_message(payment_service, message)

    def cancel_electronic_payment(
        self,
        model,      # type: eTree
        request,    # type: OrderInsertPaymentRequest
    ):
        # type: (...) -> None

        if request.electronic_type_id in self.electronic_payments_id:
            payment_service = self._get_electronic_payment_service(model=model)
            self._cancel_electronic_payment(payment_service=payment_service)

    def _cancel_electronic_payment(self, payment_service):
        # type: (str) -> None

        message = Message(token=BusTokens.TK_ELECTRONIC_PAYMENT_CANCEL_PAYMENT)
        self.message_bus.send_message(payment_service, message)

    def _get_electronic_fiscal_params(
      self,
      model,        # type: Message
      pos_id,       # type: str
      order_id,     # type: str
      payment_id,   # type: int
      amount,       # type: float
      response,     # type: Message
    ):
        # type: (...) -> Dict

        order = self.pos_repository.get_order_picture(
            pos_id=pos_id,
            order_id=order_id,
            model=model,
        )

        change_amount = "0"
        tender_seq_id = str(self._get_last_tender_seq_id(order))

        fiscal_params = self._build_fiscal_params(
            order.get("posId"),
            order.get("orderId"),
            tender_seq_id,
            str(payment_id),
            str(amount),
            change_amount
        )

        return self._build_additional_fiscal_params(fiscal_params, response)

    def _validate_electronic_payment_response(
        self,
        response,   # type: Message
        amount,     # type: float
        pos_id,     # type: str
        order_id,   # type: str

    ):
        # type: (...) -> float

        if response.token != TK_SYS_ACK or not response.data:
            raise OperationErrors(OperationErrorsType.InvalidPaymentResponse)

        response_xml = eTree.fromstring(response.data)
        result = response_xml.get("Result")
        if result != "0":
            log_message = "Error validate electronic payment: [pos_id: {} order_id: {} result: {}]"
            self.logger.info(log_message.format(pos_id, order_id, result))
            raise OperationErrors(OperationErrorsType.UnexpectedPaymentResponse)

        approved_amount = response_xml.get("ApprovedAmount").replace(",", ".")
        if not approved_amount or float(approved_amount) == float(amount):
            return amount

        message = "Approved amount [{}] is different of the order amount [{}]".format(approved_amount, amount)
        self.logger.info(message)

        return float(approved_amount)

    @staticmethod
    def _get_electronic_payment_service(model):
        # type: (eTree) -> str

        payment_service = get_used_service(model, "eft")
        if payment_service:
            return payment_service
        raise OperationErrors(OperationErrorsType.InvalidPaymentService)

    @staticmethod
    def _get_electronic_payment_data(
        model,                # type: eTree
        order,                # type: eTree
        pos_id,               # type: str
        amount,               # type: float
        electronic_type_id,   # type: str
    ):
        # type: (...) -> str

        order_id = order.get("orderId")
        session_id = model.find("Operator").get("sessionId")
        operator = model.find("Operator").get("id")
        amount = "{:.2f}".format(amount)
        electronic_type = electronic_type_id
        tender_seq_id = ''

        totaled_states = order.findall("StateHistory/State[@state='TOTALED']")
        last_totaled = totaled_states[-1].get("timestamp").replace("-", "").replace(":", "")
        last_totaled_date = last_totaled
        fiscal_date = last_totaled_date[:8]
        fiscal_time = last_totaled_date[9:15]

        data = ";".join(
            [
                pos_id,
                order_id,
                session_id,
                operator,
                electronic_type,
                amount,
                fiscal_date,
                fiscal_time,
                tender_seq_id
            ]
        )

        return data

    def _save_promotional_tenders_payment_data(self, model):
        # type: (eTree) -> None

        order = self._get_current_order(model)
        promotional_tenders = self._get_order_promotional_tenders(order)
        for tender in promotional_tenders:
            fiscal_params = self._build_fiscal_params(order.get("posId"),
                                                      order.get("orderId"),
                                                      promotional_tenders[tender].get("tender_seq_id"),
                                                      promotional_tenders[tender].get("type"),
                                                      promotional_tenders[tender].get("amount"),
                                                      promotional_tenders[tender].get("change"))

            self._store_payment_data(fiscal_params)

    @staticmethod
    def _build_fiscal_params(pos_id, order_id, tender_seq_id, tender_type, amount, change):
        # type: (str, str, str, str, str, str) -> Dict

        return {
            "pos_id": pos_id,
            "order_id": order_id,
            "tender_seq_id": tender_seq_id,
            "type": tender_type,
            "amount": amount,
            "change": change
        }

    @staticmethod
    def _build_additional_fiscal_params(fiscal_params, response):
        # type: (Dict, Message) -> Dict

        response_xml = eTree.fromstring(response.data)
        if response_xml is not None:
            fiscal_params["auth_code"] = response_xml.get("AuthCode", "")
            fiscal_params["id_auth"] = response_xml.get("IdAuth", "")
            fiscal_params["cnpj_auth"] = response_xml.get("CNPJAuth", "")
            fiscal_params["bandeira"] = response_xml.get("Bandeira", "")
            fiscal_params["t_band"] = response_xml.get("tBand", "")
            fiscal_params["t_pag"] = response_xml.get("tPag", "")
            fiscal_params["receipt_merchant"] = base64.b64decode(response_xml.get("ReceiptMerchant", ""))
            fiscal_params["receipt_customer"] = base64.b64decode(response_xml.get("ReceiptCustomer", ""))

        return fiscal_params

    @staticmethod
    def _get_order_promotional_tenders(order):
        # type: (eTree) -> Dict

        promotional_tenders = {}
        for tender in order.findall("TenderHistory/Tender"):
            tender_detail = tender.get("tenderDetail")
            if tender_detail:
                is_promotional_tender = json.loads(tender_detail).get("promotionalTender", False)
                if is_promotional_tender:
                    tender_id = tender.get("tenderId")
                    promotional_tenders[tender_id] = {
                        "amount": tender.get("tenderAmount"),
                        "tender_seq_id": tender_id,
                        "type": tender.get("tenderType"),
                        "change": "0"
                    }

        return promotional_tenders

    def _store_payment_data(self, fiscal_params):
        # type: (Dict) -> None

        def inner_func(conn):
            # type: (Connection) -> None

            conn.pquery("fiscal_savePaymentData", **fiscal_params)

        self.execute_with_transaction(inner_func, service_name="FiscalPersistence")

    def get_electronic_payments_id(self):
        # type: () -> None

        def inner_func(conn):
            # type: (Connection) -> List[int]

            query = "SELECT ElectronicType FROM TenderType WHERE ElectronicType <> 0"
            cursor = conn.select(query)

            electronic_type_ids = []
            for row in cursor:
                electronic_type_id = int(row.get_entry("ElectronicType"))
                if electronic_type_id not in electronic_type_ids:
                    electronic_type_ids.append(electronic_type_id)

            return electronic_type_ids

        self.electronic_payments_id = self.execute_with_connection(inner_func, service_name="Persistence")

    def send_electronic_payment_message_response(self, model, request):
        # type: (eTree, SendResponsesRequest) -> None

        payment_service = self._get_electronic_payment_service(model)
        data = "|".join([model.get("posId"), request.response])
        message = Message(token=BusTokens.TK_ELECTRONIC_PAYMENT_SEND_RESPONSE, data_type=DataType.param, data=data)
        response = self.message_bus.send_message(payment_service, message)
        if response.token != TK_SYS_ACK:
            raise OperationErrors(OperationErrorsType.InvalidPaymentResponse)

    @staticmethod
    def _get_last_tender_seq_id(order):
        return len(order.findall("TenderHistory/Tender")) + 1

    @staticmethod
    def _get_current_order(model):
        return model.find("CurrentOrder/Order")

    @staticmethod
    def _get_tender_details(
        payment_data  # type: Dict
    ):
        # type: (...) -> str

        tender_details = json.dumps({
            "Media": payment_data.get("Media", "").strip(),
            "CNPJAuth": payment_data.get("CNPJAuth", ""),
            "TransactionProcessor": payment_data.get("TransactionProcessor", ""),
            "Bandeira": payment_data.get("Bandeira", ""),
            "IdAuth": payment_data.get("IdAuth", ""),
            "AuthCode": payment_data.get("AuthCode", ""),
            "ReceiptCustomer": payment_data.get("ReceiptCustomer", ""),
            "ReceiptMerchant": payment_data.get("ReceiptMerchant", ""),
            "AccreditorSat": payment_data.get("AccreditorSat", ""),
            "tBand": payment_data.get("tBand", ""),
            "tPag": payment_data.get("tPag", ""),
            "NsuHost": payment_data.get("NsuHost", ""),
            "CardNumber": payment_data.get("CardNumber", ""),
            "TransactionIdentifier": payment_data.get("TransactionIdentifier"),
            "LastDigits": payment_data.get("LastDigits", ""),
        })

        return tender_details

    def get_customer_receipts(
        self,
        model,      # type: eTree
    ):
        # type: (...) -> List[str]

        order = self._get_current_order(model)
        customer_receipts = []
        for tender in order.findall("TenderHistory/Tender"):
            if int(tender.get("tenderType", -1)) in self.electronic_payments_id:
                tender_detail = json.loads(tender.get("tenderDetail", "{}"))
                tef_customer_receipt = tender_detail.get("ReceiptCustomer", '') if tender_detail is not None else None
                if tef_customer_receipt not in ['None', None, '']:
                    customer_receipts.append(tef_customer_receipt)

        return customer_receipts



from report_formatters import center, join, double_line_separator, add_amount_line, break_line, empty_line,\
                              add_table_line, add_key_value_line, DATE_TIME_FMT, single_line_separator, format_line, \
                              format_price, str_float
from default_header import report_default_header
from get_store_id import get_store_id
from get_pos_list import get_pos_list
from payment_util import is_electronic_tender_id



from ._TableReportSaleLineDto import TableReportSaleLineDto  # noqa
from typing import List  # noqa


class TableReportOrderDto(object):

    def __init__(self):
        self.order_id = None
        self.total_amount = 0.0
        self.total_after_discount = 0.0
        self.total_tender = 0.0
        self.discount_amount = 0.0
        self.sales_line_list = []  # type: List[TableReportSaleLineDto]
        self.split_key = None  # type: str
        self.tip = 0.0



# -*- coding: utf-8 -*-
from _NfceDisablerResponseParser import NfceDisablerResponseParser
from _NfceDisablerRequestBuilder import NfceDisablerRequestBuilder
from _NfceDisablerResponse import NfceDisablerResponse
from _ProcInutNfeBuilder import ProcInutNfeBuilder
from _ProcInutNfeSaver import ProcInutNfeSaver
from nfcebuilder.nfceutil import NfceRequest
from requests import exceptions


class NfceDisabler(object):
    def __init__(self,
                 nfce_disabler_request_builder,
                 nfce_disabler_response_parser,
                 nfce_request, url_inutilizacao,
                 proc_inut_nfe_builder,
                 proc_inut_nfe_saver,
                 versao_ws):
        # type: (NfceDisablerRequestBuilder, NfceDisablerResponseParser, NfceRequest, str, ProcInutNfeBuilder, ProcInutNfeSaver, int) -> None
        self.nfce_disabler_request_builder = nfce_disabler_request_builder
        self.nfce_disabler_response_parser = nfce_disabler_response_parser
        self.nfce_request = nfce_request
        self.url_inutilizacao = url_inutilizacao
        self.proc_inut_nfe_builder = proc_inut_nfe_builder
        self.proc_inut_nfe_saver = proc_inut_nfe_saver
        self.versao_ws = versao_ws

        if versao_ws == 1:
            self.soap_act = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeInutilizacao2/nfeInutilizacaoNF2"
        elif versao_ws == 3:
            self.soap_act = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeInutilizacao3/nfeInutilizacaoNF3"
        else:
            self.soap_act = "http://www.portalfiscal.inf.br/nfe/wsdl/NFeInutilizacao4/nfeInutilizacaoNF4"

    def disable_fiscal_number(self, order):
        disable_request_xml = self.nfce_disabler_request_builder.build_request(order)

        response = self.nfce_request.envia_nfce(disable_request_xml, self.url_inutilizacao, soap_action=self.soap_act)
        if response.status_code != 200:
            raise exceptions.HTTPError("Invalid response code: {0}: {1}".format(response.status_code, response.text))

        nfce_disable_response = self.nfce_disabler_response_parser.parse_response(response.text)
        xml = self.proc_inut_nfe_builder.build_proc_inut_nfe_xml(disable_request_xml, response.text, self.versao_ws)

        if nfce_disable_response.c_stat not in (
                NfceDisablerResponse.Disabled,
                NfceDisablerResponse.AlreadyDisabled,
                NfceDisablerResponse.EqualRequest,
                NfceDisablerResponse.OneOfTheRangeAlreadyDisabled):
            self.proc_inut_nfe_saver.save_proc_inut_nfe_xml_error(xml, order)
            raise Exception('Invalid cStat {0}'.format(nfce_disable_response.c_stat))

        self.proc_inut_nfe_saver.save_proc_inut_nfe_xml(xml, order)

        return xml



from report import Part
from typing import List, Union, Optional  # noqa

from ._AlignTypes import AlignTypes
from ._ReportColumnDefinition import ReportColumnDefinition
from ._I18nReport import I18nReport
from helper import remove_accents


class TableReport(I18nReport):
    def __init__(
        self,
        lines=None,                 # type: List[List[Union[str, ReportColumnDefinition]]]
        column_definitions=None,    # type: Optional[List[ReportColumnDefinition]]
        encoding='ASCII',           # type: Optional[str]
    ):

        # type: (...) -> None

        self.lines = lines
        self.column_definitions = column_definitions
        self.encoding = encoding

    def get_parts(self):
        translated_lines = TableTranslator(i18n=self.i18n).translate(
            lines=self.lines,
            encoding=self.encoding,
        )

        report = []
        for line in translated_lines:
            report.extend(self._format_line(line))
        return report

    def _format_line(self, line):
        report_line = []
        current_column = 0
        for cell in line:
            if not cell:
                continue

            line_text = self._format_cell(cell, current_column)
            report_line.append(Part(text=line_text))
            current_column += 1
        report_line.append(Part(text="\n"))
        return report_line

    def _format_cell(self, cell, current_column):
        cell_format = self._get_cell_format(cell, current_column)
        cell_text = self._get_cell_text(cell)
        return self._get_formatted_text(cell_format, cell_text)

    def _get_cell_format(self, cell, current_column):
        if isinstance(cell, ReportColumnDefinition):
            return cell
        else:
            return self.column_definitions[current_column]

    @staticmethod
    def _get_cell_text(cell):
        if isinstance(cell, ReportColumnDefinition):
            return cell.text
        else:
            return cell

    def _get_formatted_text(self, cell_format, cell_text):
        text = self._truncate_text(cell_format, cell_text)
        text = cell_format.before_text + text + cell_format.after_text
        text = self._fill_with_character(cell_format, text)

        return text

    def _truncate_text(self, cell_format, cell_text):
        reduced_cell_text = cell_text
        size_to_reduce = self.calculate_size_to_reduce(cell_format, cell_text)
        if size_to_reduce > 0:
            reduced_cell_text = cell_text[0:len(cell_text) - size_to_reduce]
        return reduced_cell_text

    @staticmethod
    def calculate_size_to_reduce(cell_format, cell_text):
        total_size = len(cell_text)
        total_size += len(cell_format.before_text) + len(cell_format.after_text)
        total_size += len(cell_format.before_fill_text) + len(cell_format.after_fill_text)
        size_to_reduce = total_size - cell_format.width
        return size_to_reduce

    def _fill_with_character(self, cell_format, text):
        fill_char = cell_format.fill_with_char
        if fill_char == "":
            fill_char = " "

        quantity = cell_format.width - len(text) - len(cell_format.after_fill_text) - len(cell_format.before_fill_text)

        if cell_format.align == AlignTypes.LEFT:
            text = self._add_fill_char_after(fill_char, quantity, text)
        elif cell_format.align == AlignTypes.RIGHT:
            text = self._add_fill_char_before(fill_char, quantity, text)
        else:
            text = self._add_fill_char_aroud(
                fill_char=fill_char,
                fill_with_char_quantity=quantity,
                text=text,
                encoding=self.encoding,
            )

        return cell_format.before_fill_text + text + cell_format.after_fill_text

    @staticmethod
    def _add_fill_char_after(fill_char, fill_with_char_quantity, text):
        return text + (fill_char * fill_with_char_quantity)

    @staticmethod
    def _add_fill_char_before(fill_char, fill_with_char_quantity, text):
        return (fill_char * fill_with_char_quantity) + text

    @staticmethod
    def _add_fill_char_aroud(
        fill_char,                  # type: str
        fill_with_char_quantity,    # type: int
        text,                       # type: str
        encoding='ASCII',           # type: Optional[str]
    ):
        # type: (...) -> str

        if encoding == 'ASCII':
            text = remove_accents(text=text)

        text = "{}{}{}".format(fill_char * (fill_with_char_quantity / 2),
                               text,
                               fill_char * (fill_with_char_quantity / 2))
        if fill_with_char_quantity % 2 == 1:
            text += fill_char
        return text


class TableTranslator(object):
    def __init__(self, i18n):
        self.i18n = i18n

    def translate(
        self,
        lines,              # type: List[List[Union[str, ReportColumnDefinition]]]
        encoding='ASCII',   # type: Optional[str]
    ):
        # type: (...) -> List

        translated_lines = []
        for line in lines:
            translated_line = []
            for cell in line:
                text = self.get_cell_text(cell)
                try:
                    translated_text = self.i18n.internationalize_text(
                        text=text,
                        encoding=encoding,
                    )
                    text = self.create_translated_cell(cell, translated_text)
                except Exception as _:
                    pass

                translated_line.append(self.create_translated_cell(cell, text))

            translated_lines.append(translated_line)

        return translated_lines

    @staticmethod
    def get_cell_text(cell):
        if isinstance(cell, ReportColumnDefinition):
            text = cell.text
        else:
            text = cell
        return text

    @staticmethod
    def create_translated_cell(cell, translated_text):
        if isinstance(cell, ReportColumnDefinition):
            return ReportColumnDefinition(text=translated_text,
                                          width=cell.width,
                                          fill_with_char=cell.fill_with_char,
                                          before_text=cell.before_text,
                                          after_text=cell.after_text,
                                          align=cell.align,
                                          after_fill_text=cell.after_fill_text,
                                          before_fill_text=cell.before_fill_text)
        else:
            return translated_text



from logging import Logger
from typing import Dict

from interactor.route.customization.get_text.dto import GetTextRequest
from model.config import Config
from model.error_type import PosValidationErrorsType
from model.exception import ValidationError
from model.repository import PosRepositoryModel
from validation_util import validate_api_key


class GetTextInteractor(object):

    def __init__(
            self,
            logger,          # type: Logger
            config,          # type: Config
            pos_repository   # type: PosRepositoryModel
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository

    def execute(
            self,
            request   # type: GetTextRequest
    ):
        # type: (...) -> Dict

        validate_api_key(request, self.config.api_key_mapping)

        text_by_language = self.pos_repository.get_text_from_i18n()
        if not text_by_language:
            raise ValidationError(PosValidationErrorsType.LanguagesNotFound)

        return text_by_language



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2001-2017 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""DNS Opcodes."""

import dns.exception

#: Query
QUERY = 0
#: Inverse Query (historical)
IQUERY = 1
#: Server Status (unspecified and unimplemented anywhere)
STATUS = 2
#: Notify
NOTIFY = 4
#: Dynamic Update
UPDATE = 5

_by_text = {
    'QUERY': QUERY,
    'IQUERY': IQUERY,
    'STATUS': STATUS,
    'NOTIFY': NOTIFY,
    'UPDATE': UPDATE
}

# We construct the inverse mapping programmatically to ensure that we
# cannot make any mistakes (e.g. omissions, cut-and-paste errors) that
# would cause the mapping not to be true inverse.

_by_value = {y: x for x, y in _by_text.items()}


class UnknownOpcode(dns.exception.DNSException):
    """An DNS opcode is unknown."""


def from_text(text):
    """Convert text into an opcode.

    *text*, a ``text``, the textual opcode

    Raises ``dns.opcode.UnknownOpcode`` if the opcode is unknown.

    Returns an ``int``.
    """

    if text.isdigit():
        value = int(text)
        if value >= 0 and value <= 15:
            return value
    value = _by_text.get(text.upper())
    if value is None:
        raise UnknownOpcode
    return value


def from_flags(flags):
    """Extract an opcode from DNS message flags.

    *flags*, an ``int``, the DNS flags.

    Returns an ``int``.
    """

    return (flags & 0x7800) >> 11


def to_flags(value):
    """Convert an opcode to a value suitable for ORing into DNS message
    flags.

    *value*, an ``int``, the DNS opcode value.

    Returns an ``int``.
    """

    return (value << 11) & 0x7800


def to_text(value):
    """Convert an opcode to text.

    *value*, an ``int`` the opcode value,

    Raises ``dns.opcode.UnknownOpcode`` if the opcode is unknown.

    Returns a ``text``.
    """

    text = _by_value.get(value)
    if text is None:
        text = str(value)
    return text


def is_update(flags):
    """Is the opcode in flags UPDATE?

    *flags*, an ``int``, the DNS message flags.

    Returns a ``bool``.
    """

    return from_flags(flags) == UPDATE



from customer_services.interactor.add_customer.dto import AddCustomerRequest
from customer_services.model import CustomerExistsException, Customer, Address, CustomerNonExistsException
from customer_services.model.repository import CustomerRepository
from customer_services.service import GetCustomerService


class AddCustomerInteractor:
    def __init__(self, customer_repository, get_customer_service):
        # type: (CustomerRepository, GetCustomerService) -> None

        self.customer_repository = customer_repository
        self.get_customer_service = get_customer_service
    
    def add_customer(self, request):
        # type: (AddCustomerRequest) -> None
        
        customer = None
        try:
            customer = self.get_customer_service.get_customer(request.phone)
        except CustomerNonExistsException as _:
            pass
        
        if customer:
            raise CustomerExistsException()
        
        address = Address(
            request.address_id,
            request.zip_code,
            request.state,
            request.city,
            request.street,
            request.neighborhood,
            request.number,
            request.complement,
            request.reference_point
        )
        customer = Customer(request.phone, request.document, request.name, address)
        self.customer_repository.add_customer(customer)



from __future__ import absolute_import
from future.utils import PY3

if PY3:
    from builtins import *
else:
    __future_module__ = True
    from __builtin__ import *
    # Overwrite any old definitions with the equivalent future.builtins ones:
    from future.builtins import *



from unicodedata import normalize

from report import Formatter
from helper import remove_accents

from _CommandFactory import CommandFactory

ESC = b'\x1b'


class EscPosFormatter(Formatter):
    def __init__(self, encoding="cp860", command_factory=None):
        # type: (str, CommandFactory) -> None
        self.encoding = encoding
        self.command_factory = command_factory

    def format(self, parts):
        body = self.create_report_body(parts)
        if body == "":
            return body
        content = ""
        try:
            content = self._get_page_command() + body
            return remove_accents(content).encode(self.encoding)
        except UnicodeDecodeError as _:
            return content

    def create_report_body(self, parts):
        body = ""
        for part in parts:
            if part.commands is not None:
                for command in part.commands:
                    body += self.command_factory.get_processor(command).get_bytes(command)
            if part.text is not None:
                body += part.text
        return body

    def _get_page_command(self):
        if self.encoding == "cp850":
            return ESC + 't' + '\x02'
        else:
            return ESC + 't' + '\x03'



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2003-2017 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Talk to a DNS server."""

from __future__ import generators

import errno
import select
import socket
import struct
import sys
import time

import dns.exception
import dns.inet
import dns.name
import dns.message
import dns.rcode
import dns.rdataclass
import dns.rdatatype
from ._compat import long, string_types, PY3

if PY3:
    select_error = OSError
else:
    select_error = select.error

# Function used to create a socket.  Can be overridden if needed in special
# situations.
socket_factory = socket.socket

class UnexpectedSource(dns.exception.DNSException):
    """A DNS query response came from an unexpected address or port."""


class BadResponse(dns.exception.FormError):
    """A DNS query response does not respond to the question asked."""


class TransferError(dns.exception.DNSException):
    """A zone transfer response got a non-zero rcode."""

    def __init__(self, rcode):
        message = 'Zone transfer error: %s' % dns.rcode.to_text(rcode)
        super(TransferError, self).__init__(message)
        self.rcode = rcode


def _compute_expiration(timeout):
    if timeout is None:
        return None
    else:
        return time.time() + timeout

# This module can use either poll() or select() as the "polling backend".
#
# A backend function takes an fd, bools for readability, writablity, and
# error detection, and a timeout.

def _poll_for(fd, readable, writable, error, timeout):
    """Poll polling backend."""

    event_mask = 0
    if readable:
        event_mask |= select.POLLIN
    if writable:
        event_mask |= select.POLLOUT
    if error:
        event_mask |= select.POLLERR

    pollable = select.poll()
    pollable.register(fd, event_mask)

    if timeout:
        event_list = pollable.poll(long(timeout * 1000))
    else:
        event_list = pollable.poll()

    return bool(event_list)


def _select_for(fd, readable, writable, error, timeout):
    """Select polling backend."""

    rset, wset, xset = [], [], []

    if readable:
        rset = [fd]
    if writable:
        wset = [fd]
    if error:
        xset = [fd]

    if timeout is None:
        (rcount, wcount, xcount) = select.select(rset, wset, xset)
    else:
        (rcount, wcount, xcount) = select.select(rset, wset, xset, timeout)

    return bool((rcount or wcount or xcount))


def _wait_for(fd, readable, writable, error, expiration):
    # Use the selected polling backend to wait for any of the specified
    # events.  An "expiration" absolute time is converted into a relative
    # timeout.

    done = False
    while not done:
        if expiration is None:
            timeout = None
        else:
            timeout = expiration - time.time()
            if timeout <= 0.0:
                raise dns.exception.Timeout
        try:
            if not _polling_backend(fd, readable, writable, error, timeout):
                raise dns.exception.Timeout
        except select_error as e:
            if e.args[0] != errno.EINTR:
                raise e
        done = True


def _set_polling_backend(fn):
    # Internal API. Do not use.

    global _polling_backend

    _polling_backend = fn

if hasattr(select, 'poll'):
    # Prefer poll() on platforms that support it because it has no
    # limits on the maximum value of a file descriptor (plus it will
    # be more efficient for high values).
    _polling_backend = _poll_for
else:
    _polling_backend = _select_for


def _wait_for_readable(s, expiration):
    _wait_for(s, True, False, True, expiration)


def _wait_for_writable(s, expiration):
    _wait_for(s, False, True, True, expiration)


def _addresses_equal(af, a1, a2):
    # Convert the first value of the tuple, which is a textual format
    # address into binary form, so that we are not confused by different
    # textual representations of the same address
    try:
        n1 = dns.inet.inet_pton(af, a1[0])
        n2 = dns.inet.inet_pton(af, a2[0])
    except dns.exception.SyntaxError:
        return False
    return n1 == n2 and a1[1:] == a2[1:]


def _destination_and_source(af, where, port, source, source_port):
    # Apply defaults and compute destination and source tuples
    # suitable for use in connect(), sendto(), or bind().
    if af is None:
        try:
            af = dns.inet.af_for_address(where)
        except Exception:
            af = dns.inet.AF_INET
    if af == dns.inet.AF_INET:
        destination = (where, port)
        if source is not None or source_port != 0:
            if source is None:
                source = '0.0.0.0'
            source = (source, source_port)
    elif af == dns.inet.AF_INET6:
        destination = (where, port, 0, 0)
        if source is not None or source_port != 0:
            if source is None:
                source = '::'
            source = (source, source_port, 0, 0)
    return (af, destination, source)


def send_udp(sock, what, destination, expiration=None):
    """Send a DNS message to the specified UDP socket.

    *sock*, a ``socket``.

    *what*, a ``binary`` or ``dns.message.Message``, the message to send.

    *destination*, a destination tuple appropriate for the address family
    of the socket, specifying where to send the query.

    *expiration*, a ``float`` or ``None``, the absolute time at which
    a timeout exception should be raised.  If ``None``, no timeout will
    occur.

    Returns an ``(int, float)`` tuple of bytes sent and the sent time.
    """

    if isinstance(what, dns.message.Message):
        what = what.to_wire()
    _wait_for_writable(sock, expiration)
    sent_time = time.time()
    n = sock.sendto(what, destination)
    return (n, sent_time)


def receive_udp(sock, destination, expiration=None,
                ignore_unexpected=False, one_rr_per_rrset=False,
                keyring=None, request_mac=b'', ignore_trailing=False):
    """Read a DNS message from a UDP socket.

    *sock*, a ``socket``.

    *destination*, a destination tuple appropriate for the address family
    of the socket, specifying where the associated query was sent.

    *expiration*, a ``float`` or ``None``, the absolute time at which
    a timeout exception should be raised.  If ``None``, no timeout will
    occur.

    *ignore_unexpected*, a ``bool``.  If ``True``, ignore responses from
    unexpected sources.

    *one_rr_per_rrset*, a ``bool``.  If ``True``, put each RR into its own
    RRset.

    *keyring*, a ``dict``, the keyring to use for TSIG.

    *request_mac*, a ``binary``, the MAC of the request (for TSIG).

    *ignore_trailing*, a ``bool``.  If ``True``, ignore trailing
    junk at end of the received message.

    Raises if the message is malformed, if network errors occur, of if
    there is a timeout.

    Returns a ``dns.message.Message`` object.
    """

    wire = b''
    while 1:
        _wait_for_readable(sock, expiration)
        (wire, from_address) = sock.recvfrom(65535)
        if _addresses_equal(sock.family, from_address, destination) or \
           (dns.inet.is_multicast(destination[0]) and
            from_address[1:] == destination[1:]):
            break
        if not ignore_unexpected:
            raise UnexpectedSource('got a response from '
                                   '%s instead of %s' % (from_address,
                                                         destination))
    received_time = time.time()
    r = dns.message.from_wire(wire, keyring=keyring, request_mac=request_mac,
                              one_rr_per_rrset=one_rr_per_rrset,
                              ignore_trailing=ignore_trailing)
    return (r, received_time)

def udp(q, where, timeout=None, port=53, af=None, source=None, source_port=0,
        ignore_unexpected=False, one_rr_per_rrset=False, ignore_trailing=False):
    """Return the response obtained after sending a query via UDP.

    *q*, a ``dns.message.Message``, the query to send

    *where*, a ``text`` containing an IPv4 or IPv6 address,  where
    to send the message.

    *timeout*, a ``float`` or ``None``, the number of seconds to wait before the
    query times out.  If ``None``, the default, wait forever.

    *port*, an ``int``, the port send the message to.  The default is 53.

    *af*, an ``int``, the address family to use.  The default is ``None``,
    which causes the address family to use to be inferred from the form of
    *where*.  If the inference attempt fails, AF_INET is used.  This
    parameter is historical; you need never set it.

    *source*, a ``text`` containing an IPv4 or IPv6 address, specifying
    the source address.  The default is the wildcard address.

    *source_port*, an ``int``, the port from which to send the message.
    The default is 0.

    *ignore_unexpected*, a ``bool``.  If ``True``, ignore responses from
    unexpected sources.

    *one_rr_per_rrset*, a ``bool``.  If ``True``, put each RR into its own
    RRset.

    *ignore_trailing*, a ``bool``.  If ``True``, ignore trailing
    junk at end of the received message.

    Returns a ``dns.message.Message``.
    """

    wire = q.to_wire()
    (af, destination, source) = _destination_and_source(af, where, port,
                                                        source, source_port)
    s = socket_factory(af, socket.SOCK_DGRAM, 0)
    received_time = None
    sent_time = None
    try:
        expiration = _compute_expiration(timeout)
        s.setblocking(0)
        if source is not None:
            s.bind(source)
        (_, sent_time) = send_udp(s, wire, destination, expiration)
        (r, received_time) = receive_udp(s, destination, expiration,
                                         ignore_unexpected, one_rr_per_rrset,
                                         q.keyring, q.mac, ignore_trailing)
    finally:
        if sent_time is None or received_time is None:
            response_time = 0
        else:
            response_time = received_time - sent_time
        s.close()
    r.time = response_time
    if not q.is_response(r):
        raise BadResponse
    return r


def _net_read(sock, count, expiration):
    """Read the specified number of bytes from sock.  Keep trying until we
    either get the desired amount, or we hit EOF.
    A Timeout exception will be raised if the operation is not completed
    by the expiration time.
    """
    s = b''
    while count > 0:
        _wait_for_readable(sock, expiration)
        n = sock.recv(count)
        if n == b'':
            raise EOFError
        count = count - len(n)
        s = s + n
    return s


def _net_write(sock, data, expiration):
    """Write the specified data to the socket.
    A Timeout exception will be raised if the operation is not completed
    by the expiration time.
    """
    current = 0
    l = len(data)
    while current < l:
        _wait_for_writable(sock, expiration)
        current += sock.send(data[current:])


def send_tcp(sock, what, expiration=None):
    """Send a DNS message to the specified TCP socket.

    *sock*, a ``socket``.

    *what*, a ``binary`` or ``dns.message.Message``, the message to send.

    *expiration*, a ``float`` or ``None``, the absolute time at which
    a timeout exception should be raised.  If ``None``, no timeout will
    occur.

    Returns an ``(int, float)`` tuple of bytes sent and the sent time.
    """

    if isinstance(what, dns.message.Message):
        what = what.to_wire()
    l = len(what)
    # copying the wire into tcpmsg is inefficient, but lets us
    # avoid writev() or doing a short write that would get pushed
    # onto the net
    tcpmsg = struct.pack("!H", l) + what
    _wait_for_writable(sock, expiration)
    sent_time = time.time()
    _net_write(sock, tcpmsg, expiration)
    return (len(tcpmsg), sent_time)

def receive_tcp(sock, expiration=None, one_rr_per_rrset=False,
                keyring=None, request_mac=b'', ignore_trailing=False):
    """Read a DNS message from a TCP socket.

    *sock*, a ``socket``.

    *expiration*, a ``float`` or ``None``, the absolute time at which
    a timeout exception should be raised.  If ``None``, no timeout will
    occur.

    *one_rr_per_rrset*, a ``bool``.  If ``True``, put each RR into its own
    RRset.

    *keyring*, a ``dict``, the keyring to use for TSIG.

    *request_mac*, a ``binary``, the MAC of the request (for TSIG).

    *ignore_trailing*, a ``bool``.  If ``True``, ignore trailing
    junk at end of the received message.

    Raises if the message is malformed, if network errors occur, of if
    there is a timeout.

    Returns a ``dns.message.Message`` object.
    """

    ldata = _net_read(sock, 2, expiration)
    (l,) = struct.unpack("!H", ldata)
    wire = _net_read(sock, l, expiration)
    received_time = time.time()
    r = dns.message.from_wire(wire, keyring=keyring, request_mac=request_mac,
                              one_rr_per_rrset=one_rr_per_rrset,
                              ignore_trailing=ignore_trailing)
    return (r, received_time)

def _connect(s, address):
    try:
        s.connect(address)
    except socket.error:
        (ty, v) = sys.exc_info()[:2]

        if hasattr(v, 'errno'):
            v_err = v.errno
        else:
            v_err = v[0]
        if v_err not in [errno.EINPROGRESS, errno.EWOULDBLOCK, errno.EALREADY]:
            raise v


def tcp(q, where, timeout=None, port=53, af=None, source=None, source_port=0,
        one_rr_per_rrset=False, ignore_trailing=False):
    """Return the response obtained after sending a query via TCP.

    *q*, a ``dns.message.Message``, the query to send

    *where*, a ``text`` containing an IPv4 or IPv6 address,  where
    to send the message.

    *timeout*, a ``float`` or ``None``, the number of seconds to wait before the
    query times out.  If ``None``, the default, wait forever.

    *port*, an ``int``, the port send the message to.  The default is 53.

    *af*, an ``int``, the address family to use.  The default is ``None``,
    which causes the address family to use to be inferred from the form of
    *where*.  If the inference attempt fails, AF_INET is used.  This
    parameter is historical; you need never set it.

    *source*, a ``text`` containing an IPv4 or IPv6 address, specifying
    the source address.  The default is the wildcard address.

    *source_port*, an ``int``, the port from which to send the message.
    The default is 0.

    *one_rr_per_rrset*, a ``bool``.  If ``True``, put each RR into its own
    RRset.

    *ignore_trailing*, a ``bool``.  If ``True``, ignore trailing
    junk at end of the received message.

    Returns a ``dns.message.Message``.
    """

    wire = q.to_wire()
    (af, destination, source) = _destination_and_source(af, where, port,
                                                        source, source_port)
    s = socket_factory(af, socket.SOCK_STREAM, 0)
    begin_time = None
    received_time = None
    try:
        expiration = _compute_expiration(timeout)
        s.setblocking(0)
        begin_time = time.time()
        if source is not None:
            s.bind(source)
        _connect(s, destination)
        send_tcp(s, wire, expiration)
        (r, received_time) = receive_tcp(s, expiration, one_rr_per_rrset,
                                         q.keyring, q.mac, ignore_trailing)
    finally:
        if begin_time is None or received_time is None:
            response_time = 0
        else:
            response_time = received_time - begin_time
        s.close()
    r.time = response_time
    if not q.is_response(r):
        raise BadResponse
    return r


def xfr(where, zone, rdtype=dns.rdatatype.AXFR, rdclass=dns.rdataclass.IN,
        timeout=None, port=53, keyring=None, keyname=None, relativize=True,
        af=None, lifetime=None, source=None, source_port=0, serial=0,
        use_udp=False, keyalgorithm=dns.tsig.default_algorithm):
    """Return a generator for the responses to a zone transfer.

    *where*.  If the inference attempt fails, AF_INET is used.  This
    parameter is historical; you need never set it.

    *zone*, a ``dns.name.Name`` or ``text``, the name of the zone to transfer.

    *rdtype*, an ``int`` or ``text``, the type of zone transfer.  The
    default is ``dns.rdatatype.AXFR``.  ``dns.rdatatype.IXFR`` can be
    used to do an incremental transfer instead.

    *rdclass*, an ``int`` or ``text``, the class of the zone transfer.
    The default is ``dns.rdataclass.IN``.

    *timeout*, a ``float``, the number of seconds to wait for each
    response message.  If None, the default, wait forever.

    *port*, an ``int``, the port send the message to.  The default is 53.

    *keyring*, a ``dict``, the keyring to use for TSIG.

    *keyname*, a ``dns.name.Name`` or ``text``, the name of the TSIG
    key to use.

    *relativize*, a ``bool``.  If ``True``, all names in the zone will be
    relativized to the zone origin.  It is essential that the
    relativize setting matches the one specified to
    ``dns.zone.from_xfr()`` if using this generator to make a zone.

    *af*, an ``int``, the address family to use.  The default is ``None``,
    which causes the address family to use to be inferred from the form of
    *where*.  If the inference attempt fails, AF_INET is used.  This
    parameter is historical; you need never set it.

    *lifetime*, a ``float``, the total number of seconds to spend
    doing the transfer.  If ``None``, the default, then there is no
    limit on the time the transfer may take.

    *source*, a ``text`` containing an IPv4 or IPv6 address, specifying
    the source address.  The default is the wildcard address.

    *source_port*, an ``int``, the port from which to send the message.
    The default is 0.

    *serial*, an ``int``, the SOA serial number to use as the base for
    an IXFR diff sequence (only meaningful if *rdtype* is
    ``dns.rdatatype.IXFR``).

    *use_udp*, a ``bool``.  If ``True``, use UDP (only meaningful for IXFR).

    *keyalgorithm*, a ``dns.name.Name`` or ``text``, the TSIG algorithm to use.

    Raises on errors, and so does the generator.

    Returns a generator of ``dns.message.Message`` objects.
    """

    if isinstance(zone, string_types):
        zone = dns.name.from_text(zone)
    if isinstance(rdtype, string_types):
        rdtype = dns.rdatatype.from_text(rdtype)
    q = dns.message.make_query(zone, rdtype, rdclass)
    if rdtype == dns.rdatatype.IXFR:
        rrset = dns.rrset.from_text(zone, 0, 'IN', 'SOA',
                                    '. . %u 0 0 0 0' % serial)
        q.authority.append(rrset)
    if keyring is not None:
        q.use_tsig(keyring, keyname, algorithm=keyalgorithm)
    wire = q.to_wire()
    (af, destination, source) = _destination_and_source(af, where, port,
                                                        source, source_port)
    if use_udp:
        if rdtype != dns.rdatatype.IXFR:
            raise ValueError('cannot do a UDP AXFR')
        s = socket_factory(af, socket.SOCK_DGRAM, 0)
    else:
        s = socket_factory(af, socket.SOCK_STREAM, 0)
    s.setblocking(0)
    if source is not None:
        s.bind(source)
    expiration = _compute_expiration(lifetime)
    _connect(s, destination)
    l = len(wire)
    if use_udp:
        _wait_for_writable(s, expiration)
        s.send(wire)
    else:
        tcpmsg = struct.pack("!H", l) + wire
        _net_write(s, tcpmsg, expiration)
    done = False
    delete_mode = True
    expecting_SOA = False
    soa_rrset = None
    if relativize:
        origin = zone
        oname = dns.name.empty
    else:
        origin = None
        oname = zone
    tsig_ctx = None
    first = True
    while not done:
        mexpiration = _compute_expiration(timeout)
        if mexpiration is None or mexpiration > expiration:
            mexpiration = expiration
        if use_udp:
            _wait_for_readable(s, expiration)
            (wire, from_address) = s.recvfrom(65535)
        else:
            ldata = _net_read(s, 2, mexpiration)
            (l,) = struct.unpack("!H", ldata)
            wire = _net_read(s, l, mexpiration)
        is_ixfr = (rdtype == dns.rdatatype.IXFR)
        r = dns.message.from_wire(wire, keyring=q.keyring, request_mac=q.mac,
                                  xfr=True, origin=origin, tsig_ctx=tsig_ctx,
                                  multi=True, first=first,
                                  one_rr_per_rrset=is_ixfr)
        rcode = r.rcode()
        if rcode != dns.rcode.NOERROR:
            raise TransferError(rcode)
        tsig_ctx = r.tsig_ctx
        first = False
        answer_index = 0
        if soa_rrset is None:
            if not r.answer or r.answer[0].name != oname:
                raise dns.exception.FormError(
                    "No answer or RRset not for qname")
            rrset = r.answer[0]
            if rrset.rdtype != dns.rdatatype.SOA:
                raise dns.exception.FormError("first RRset is not an SOA")
            answer_index = 1
            soa_rrset = rrset.copy()
            if rdtype == dns.rdatatype.IXFR:
                if soa_rrset[0].serial <= serial:
                    #
                    # We're already up-to-date.
                    #
                    done = True
                else:
                    expecting_SOA = True
        #
        # Process SOAs in the answer section (other than the initial
        # SOA in the first message).
        #
        for rrset in r.answer[answer_index:]:
            if done:
                raise dns.exception.FormError("answers after final SOA")
            if rrset.rdtype == dns.rdatatype.SOA and rrset.name == oname:
                if expecting_SOA:
                    if rrset[0].serial != serial:
                        raise dns.exception.FormError(
                            "IXFR base serial mismatch")
                    expecting_SOA = False
                elif rdtype == dns.rdatatype.IXFR:
                    delete_mode = not delete_mode
                #
                # If this SOA RRset is equal to the first we saw then we're
                # finished. If this is an IXFR we also check that we're seeing
                # the record in the expected part of the response.
                #
                if rrset == soa_rrset and \
                        (rdtype == dns.rdatatype.AXFR or
                         (rdtype == dns.rdatatype.IXFR and delete_mode)):
                    done = True
            elif expecting_SOA:
                #
                # We made an IXFR request and are expecting another
                # SOA RR, but saw something else, so this must be an
                # AXFR response.
                #
                rdtype = dns.rdatatype.AXFR
                expecting_SOA = False
        if done and q.keyring and not r.had_tsig:
            raise dns.exception.FormError("missing TSIG")
        yield r
    s.close()
